import datetime as dt
import glob
import logging
import os.path
import re

import netCDF4
import numpy as np
import pandas as pd
import pytest
from click.testing import CliRunner
from pymagicc.io import MAGICCData
from scmdata import ScmDataFrame

from netcdf_scm.cli import stitch_netcdf_scm_ncs
from netcdf_scm.io import load_scmdataframe
from netcdf_scm.normalisation import get_normaliser
from netcdf_scm.stitching import (
    get_branch_time,
    get_continuous_timeseries_with_meta,
    get_parent_file_path,
    get_parent_replacements,
)


def _do_generic_stitched_data_tests(stiched_scmdf, cmip5=False):
    assert stiched_scmdf["scenario"].nunique() == 1
    if cmip5:
        assert "(child) branch_time" in stiched_scmdf.metadata
        assert "(parent) tracking_id" in stiched_scmdf.metadata

    else:
        assert "(child) branch_time_in_parent" in stiched_scmdf.metadata
        assert "(parent) source_id" in stiched_scmdf.metadata

    assert "(child) parent_experiment_id" in stiched_scmdf.metadata
    assert "(parent) experiment_id" in stiched_scmdf.metadata


def _do_generic_normalised_data_tests(normalised_scmdf, cmip5=False):
    assert normalised_scmdf["scenario"].nunique() == 1
    if cmip5:
        assert "(child) branch_time" in normalised_scmdf.metadata
        assert "(normalisation) tracking_id" in normalised_scmdf.metadata

    else:
        assert "(child) branch_time_in_parent" in normalised_scmdf.metadata
        assert "(normalisation) source_id" in normalised_scmdf.metadata

    assert "(child) parent_experiment_id" in normalised_scmdf.metadata
    assert "normalisation method" in normalised_scmdf.metadata
    assert "(normalisation) experiment_id" in normalised_scmdf.metadata


def _get_ref_values(
    res, normalisation, parent, normalisation_path, normalisation_method
):
    helper = res.timeseries()
    helper = pd.concat([helper.iloc[:, 0]] * res["time"].shape[0], axis=1)
    helper.columns = res["time"]

    norm_helper = normalisation.timeseries()

    useful_idx = set(norm_helper.index.names).intersection(set(helper.index.names))
    helper.index = helper.index.droplevel(
        list(set(helper.index.names) - set(useful_idx))
    )
    norm_helper.index = norm_helper.index.droplevel(
        list(set(norm_helper.index.names) - set(useful_idx))
    )

    helper = ScmDataFrame(helper)
    norm_helper = ScmDataFrame(norm_helper)
    norm_helper.metadata = normalisation.metadata

    normaliser = get_normaliser(normalisation_method)
    ref_values = normaliser.get_reference_values(
        helper, norm_helper, get_branch_time(parent, parent_path=normalisation_path),
    )

    return ref_values


def _check_expected_values(
    res, other, years, regions=None, ref_values=None, operation=None
):
    if regions is None:
        regions = res.get_unique_meta("region")

    for region in regions:
        res_vals = res.filter(region=region, year=years).values
        assert not any([s == 0 for s in res_vals.shape])

        other_ts = other.filter(region=region, year=years).timeseries()
        other_vals = other_ts.values
        assert not any([s == 0 for s in other_vals.shape])

        if ref_values is not None:
            rvs = ref_values.loc[ref_values.index.get_level_values("region") == region]
            # precision of times in a `.MAG` file is only year-month so update before
            # retrieving values of interest
            rvs.columns = rvs.columns.map(
                lambda x: "{:04d}{:02d}".format(x.year, x.month)
            )
            other_ts_year_month_cols = other_ts.columns.map(
                lambda x: "{:04d}{:02d}".format(x.year, x.month)
            )

            other_vals = other_vals - rvs[other_ts_year_month_cols].values

        if operation is not None:
            if operation == "time_mean_ac":
                other_ts.iloc[:, :] = other_vals
                other_vals = ScmDataFrame(other_ts).time_mean("AC").timeseries().values

            else:
                raise NotImplementedError("operation: {}".format(operation))

        np.testing.assert_allclose(
            res_vals, other_vals, rtol=1e-5, atol=1e-6,
        )


def _do_filepath_time_consistency_check(res, filepath, annual=False):
    if annual:
        expected_end = "{:04d}-{:04d}.MAG".format(
            res["time"].min().year, res["time"].max().year,
        )
    else:
        expected_end = "{:04d}{:02d}-{:04d}{:02d}.MAG".format(
            res["time"].min().year,
            res["time"].min().month,
            res["time"].max().year,
            res["time"].max().month,
        )
    assert filepath.endswith(expected_end)


def test_stitching_default(tmpdir, caplog, test_cmip6_crunch_output):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_default"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*EC-Earth3-Veg.*ssp585.*r1i1p1f1.*hfds.*",
            ],
        )

    assert result.exit_code == 0, result.stderr

    out_files = glob.glob(os.path.join(output_dir, "**", "*.MAG"), recursive=True)
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])
    _do_generic_stitched_data_tests(res)
    _do_filepath_time_consistency_check(res, out_files[0])

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "ssp585" in child_path
    assert "r1i1p1f1" in child_path
    assert "EC-Earth3-Veg" in child_path
    assert "hfds" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", child_path))[0]
    )

    _check_expected_values(res, child, range(2015, 2017))

    parent_path = res.metadata["(parent) netcdf-scm crunched file"]
    assert "historical" in parent_path
    assert "r1i1p1f1" in parent_path
    assert "EC-Earth3-Veg" in parent_path
    assert "hfds" in parent_path

    parent = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", parent_path))[0]
    )

    _check_expected_values(res, parent, range(2013, 2015))


def test_stitching_two_levels(tmpdir, caplog, test_cmip6_crunch_output):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_two_levels"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                "^(?!.*(NCAR)).*GeoMIP.*G6solar.*$",
            ],
        )

    assert result.exit_code == 0, result.stderr

    out_files = glob.glob(os.path.join(output_dir, "**", "*.MAG"), recursive=True)
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    _do_generic_stitched_data_tests(res)
    _do_filepath_time_consistency_check(res, out_files[0])

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "G6solar" in child_path
    assert "r1i1p1f2" in child_path
    assert "UKESM1-0-LL" in child_path
    assert "v20191031" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", child_path))[0]
    )

    _check_expected_values(
        res,
        child,
        range(2020, 2100),
        regions=["World", "World|Northern Hemisphere", "World|Land"],
    )

    parent_path = res.metadata["(parent) netcdf-scm crunched file"]
    assert "ssp585" in parent_path
    assert "r1i1p1f2" in parent_path
    assert "UKESM1-0-LL" in parent_path
    assert "v20190726" in parent_path

    parent = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", parent_path))[0]
    )

    _check_expected_values(
        res,
        parent,
        range(2015, 2020),
        regions=["World", "World|Northern Hemisphere", "World|Land"],
    )

    grandparent_path = res.metadata["(grandparent) netcdf-scm crunched file"]
    assert "historical" in grandparent_path
    assert "r1i1p1f2" in grandparent_path
    assert "UKESM1-0-LL" in grandparent_path
    assert "v20190627" in grandparent_path

    grandparent = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", grandparent_path))[0]
    )

    _check_expected_values(
        res,
        grandparent,
        range(1880, 2014),
        regions=["World", "World|Northern Hemisphere", "World|Land"],
    )


def test_stitching_in_file_BCC_CSM2_MR(tmpdir, caplog, test_cmip6_crunch_output):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_in_file_BCC_CSM2_MR"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*BCC-CSM2-MR.*ssp126.*r1i1p1f1.*tas.*",
                "--out-format",
                "mag-files-average-year-mid-year",
            ],
        )

    assert result.exit_code == 0, result.stderr

    out_files = glob.glob(os.path.join(output_dir, "**", "*.MAG"), recursive=True)
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    _do_generic_stitched_data_tests(res)
    _do_filepath_time_consistency_check(res, out_files[0], annual=True)

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "ssp126" in child_path
    assert "r1i1p1f1" in child_path
    assert "BCC-CSM2-MR" in child_path
    assert "tas" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", child_path))[0]
    )

    _check_expected_values(res, child, range(2015, 2017), operation="time_mean_ac")

    parent_path = res.metadata["(parent) netcdf-scm crunched file"]
    assert "historical" in parent_path
    assert "r1i1p1f1" in parent_path
    assert "BCC-CSM2-MR" in parent_path
    assert "tas" in parent_path

    parent = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", parent_path))[0]
    )

    _check_expected_values(res, parent, range(2013, 2015), operation="time_mean_ac")

    warn_str = (
        "Assuming BCC metadata is wrong and branch time units are actually years, "
        "not days"
    )
    bcc_warning = [r for r in caplog.record_tuples if r[2] == warn_str]
    assert len(bcc_warning) == 1
    assert bcc_warning[0][1] == logging.WARNING


def test_stitching_no_parent(tmpdir, caplog, test_cmip6_crunch_output):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_no_parent"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*CNRM-ESM2-1.*r2i1p1f2.*/cSoil/.*",
            ],
        )

    assert result.exit_code != 0

    error_msg = re.compile(
        ".*No parent data \\(ssp585\\) available for "
        ".*CMIP6/ScenarioMIP/CNRM-CERFACS/CNRM-ESM2-1/ssp534-over/r2i1p1f2/Emon/cSoil/gr/v20190410/netcdf-scm_cSoil_Emon_CNRM-ESM2-1_ssp534-over_r2i1p1f2_gr_201501-210012.nc"
        ", we looked in "
        ".*CMIP6/ScenarioMIP/CNRM-CERFACS/CNRM-ESM2-1/ssp585/r2i1p1f2/Emon/cSoil/gr/\\*/netcdf-scm_cSoil_Emon_CNRM-ESM2-1_ssp585_r2i1p1f2_gr_\\*.nc"
    )
    no_parent_error = [r for r in caplog.record_tuples if error_msg.match(r[2])]
    assert len(no_parent_error) == 1
    assert no_parent_error[0][1] == logging.ERROR


def test_stitching_historical_only(tmpdir, caplog, test_cmip6_crunch_output):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_historical_only"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*GFDL-CM4.*1pctCO2.*r1i1p1f1.*",
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(os.path.join(output_dir, "**", "*.MAG"), recursive=True)
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    # should just be same as wrangle if like this
    assert not any(["(child)" in k for k in res.metadata])
    error_msg = re.compile(
        ".*No normalisation is being done and the parent of 1pctCO2 is piControl for "
        "infile: "
        ".*CMIP6/CMIP/NOAA-GFDL/GFDL-CM4/1pctCO2/r1i1p1f1/Amon/tas/gr1/v20180701/netcdf-scm_tas_Amon_GFDL-CM4_1pctCO2_r1i1p1f1_gr1_000101-015012.nc"
    )
    no_operation_info = [r for r in caplog.record_tuples if error_msg.match(r[2])]
    assert len(no_operation_info) == 1
    assert no_operation_info[0][1] == logging.INFO


def test_stitching_esm_hist_only(tmpdir, caplog, test_cmip6_crunch_output):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_esm_hist_only"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*CanESM5.*esm-hist.*r1i1p1f1.*fgco2.*",
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(os.path.join(output_dir, "**", "*.MAG"), recursive=True)
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    # should just be same as wrangle if like this
    assert not any(["(child)" in k for k in res.metadata])
    error_msg = re.compile(
        "No normalisation is being done and the parent of esm-hist is esm-piControl "
        "for infile: "
        ".*CMIP6/CMIP/CCCma/CanESM5/esm-hist/r1i1p1f1/Omon/fgco2/gn/v20190429/netcdf-scm_fgco2_Omon_CanESM5_esm-hist_r1i1p1f1_gn_185001-201412.nc"
    )
    no_operation_info = [r for r in caplog.record_tuples if error_msg.match(r[2])]
    assert len(no_operation_info) == 1
    assert no_operation_info[0][1] == logging.INFO


def test_stitching_with_normalisation(
    tmpdir, caplog, test_cmip6_crunch_output, normalisation_method
):
    if normalisation_method.startswith("21-yr-running-mean"):
        pytest.xfail("Not enough data to take running mean")

    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*CESM2.*r10i1p1f1.*tas.*",
                "--normalise",
                normalisation_method,
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(os.path.join(output_dir, "**", "*.MAG"), recursive=True)
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    _do_generic_normalised_data_tests(res)
    _do_filepath_time_consistency_check(res, out_files[0])
    assert res.metadata["normalisation method"] == normalisation_method

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "historical" in child_path
    assert "r10i1p1f1" in child_path
    assert "CESM2" in child_path
    assert "tas" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", child_path))[0]
    )

    normalisation_path = res.metadata["(normalisation) netcdf-scm crunched file"]
    assert "piControl" in normalisation_path
    assert "r1i1p1f1" in normalisation_path
    assert "CESM2" in normalisation_path
    assert "tas" in normalisation_path

    normalisation_path = glob.glob(
        os.path.join(test_cmip6_crunch_output, "*", normalisation_path)
    )[0]
    normalisation = load_scmdataframe(normalisation_path)

    assert child.metadata["branch_time_in_parent"] == 306600.0
    assert child.metadata["parent_time_units"] == "days since 0001-01-01 00:00:00"

    ref_values = _get_ref_values(
        res, normalisation, child, normalisation_path, normalisation_method
    )

    _check_expected_values(res, child, range(1850, 2015), ref_values=ref_values)


def test_stitching_with_normalisation_esm_hist(
    tmpdir, caplog, test_cmip6_crunch_output, normalisation_method
):
    if normalisation_method.startswith("21-yr-running-mean"):
        pytest.xfail("Not enough data to take running mean")

    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation_esm_hist"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*CanESM5.*esm-hist.*r1i1p1f1.*fgco2.*",
                "--normalise",
                normalisation_method,
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(os.path.join(output_dir, "**", "*.MAG"), recursive=True)
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    _do_generic_normalised_data_tests(res)
    _do_filepath_time_consistency_check(res, out_files[0])
    assert res.metadata["normalisation method"] == normalisation_method

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "esm-hist" in child_path
    assert "r1i1p1f1" in child_path
    assert "CanESM5" in child_path
    assert "fgco2" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", child_path))[0]
    )

    normalisation_path = res.metadata["(normalisation) netcdf-scm crunched file"]
    assert "esm-piControl" in normalisation_path
    assert "r1i1p1f1" in normalisation_path
    assert "CanESM5" in normalisation_path
    assert "fgco2" in normalisation_path

    normalisation_path = glob.glob(
        os.path.join(test_cmip6_crunch_output, "*", normalisation_path)
    )[0]
    normalisation = load_scmdataframe(normalisation_path)

    assert child.metadata["branch_time_in_parent"] == 1277865.0
    assert child.metadata["parent_time_units"] == "days since 1850-01-01 0:0:0.0"

    ref_values = _get_ref_values(
        res, normalisation, child, normalisation_path, normalisation_method
    )

    _check_expected_values(res, child, range(1850, 2015), ref_values=ref_values)


def test_stitching_with_normalisation_year_zero(
    tmpdir, caplog, test_cmip6_crunch_output, normalisation_method
):
    if normalisation_method.startswith("21-yr-running-mean"):
        pytest.xfail("Not enough data to take running mean")

    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation_year_zero"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*CESM2.*abrupt-4xCO2.*r1i1p1f1.*fgco2.*",
                "--normalise",
                normalisation_method,
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(os.path.join(output_dir, "**", "*.MAG"), recursive=True)
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    _do_generic_normalised_data_tests(res)
    _do_filepath_time_consistency_check(res, out_files[0])
    assert res.metadata["normalisation method"] == normalisation_method

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "abrupt-4xCO2" in child_path
    assert "r1i1p1f1" in child_path
    assert "CESM2" in child_path
    assert "fgco2" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", child_path))[0]
    )

    normalisation_path = res.metadata["(normalisation) netcdf-scm crunched file"]
    assert "piControl" in normalisation_path
    assert "r1i1p1f1" in normalisation_path
    assert "CESM2" in normalisation_path
    assert "fgco2" in normalisation_path

    normalisation_path = glob.glob(
        os.path.join(test_cmip6_crunch_output, "*", normalisation_path)
    )[0]
    normalisation = load_scmdataframe(normalisation_path)

    assert child.metadata["branch_time_in_parent"] == 182865.0
    assert child.metadata["parent_time_units"] == "days since 0001-01-01 00:00:00"

    ref_values = _get_ref_values(
        res, normalisation, child, normalisation_path, normalisation_method
    )

    _check_expected_values(res, child, range(1, 1000), ref_values=ref_values)


def test_stitching_with_normalisation_two_levels(
    tmpdir, caplog, test_cmip6_crunch_output, normalisation_method
):
    if normalisation_method.startswith("21-yr-running-mean"):
        pytest.xfail("Not enough data to take running mean")

    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation_two_levels"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*BCC-CSM2-MR.*ssp126.*tas.*",
                "--normalise",
                normalisation_method,
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(os.path.join(output_dir, "**", "*.MAG"), recursive=True)
    assert len(out_files) == 1

    warn_str = (
        "Assuming BCC metadata is wrong and branch time units are actually years, "
        "not days"
    )
    bcc_warning = [r for r in caplog.record_tuples if r[2] == warn_str]
    assert len(bcc_warning) == 2
    assert bcc_warning[0][1] == logging.WARNING

    res = MAGICCData(out_files[0])

    _do_generic_stitched_data_tests(res)
    _do_generic_normalised_data_tests(res)
    _do_filepath_time_consistency_check(res, out_files[0])
    assert res.metadata["normalisation method"] == normalisation_method

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "ssp126" in child_path
    assert "r1i1p1f1" in child_path
    assert "BCC-CSM2-MR" in child_path
    assert "tas" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", child_path))[0]
    )

    parent_path = res.metadata["(parent) netcdf-scm crunched file"]
    assert "historical" in parent_path
    assert "r1i1p1f1" in child_path
    assert "BCC-CSM2-MR" in child_path
    assert "tas" in child_path

    parent = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", parent_path))[0]
    )

    normalisation_path = res.metadata["(normalisation) netcdf-scm crunched file"]
    assert "piControl" in normalisation_path
    assert "r1i1p1f1" in normalisation_path
    assert "BCC-CSM2-MR" in normalisation_path
    assert "tas" in normalisation_path

    normalisation_path = glob.glob(
        os.path.join(test_cmip6_crunch_output, "*", normalisation_path)
    )[0]
    normalisation = load_scmdataframe(normalisation_path)

    assert parent.metadata["branch_time_in_parent"] == 2289.0
    assert parent.metadata["parent_time_units"] == "days since 1850-01-01"

    ref_values = _get_ref_values(
        res, normalisation, parent, normalisation_path, normalisation_method
    )

    _check_expected_values(res, child, range(2015, 2101), ref_values=ref_values)

    _check_expected_values(res, parent, range(1850, 2015), ref_values=ref_values)


def test_stitching_with_normalisation_two_levels_cmip5(
    tmpdir, caplog, test_marble_cmip5_crunch_output, normalisation_method
):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation_two_levels_cmip5"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_marble_cmip5_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "MarbleCMIP5",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*rcp45.*tas.*NorESM1.*",
                "--normalise",
                normalisation_method,
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(os.path.join(output_dir, "**", "*.MAG"), recursive=True)
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    _do_generic_stitched_data_tests(res, cmip5=True)
    _do_generic_normalised_data_tests(res, cmip5=True)
    _do_filepath_time_consistency_check(res, out_files[0])
    assert res.metadata["normalisation method"] == normalisation_method

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "rcp45" in child_path
    assert "r1i1p1" in child_path
    assert "NorESM1-M" in child_path
    assert "tas" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_marble_cmip5_crunch_output, "*", child_path))[0]
    )

    parent_path = res.metadata["(parent) netcdf-scm crunched file"]
    assert "historical" in parent_path
    assert "r1i1p1" in child_path
    assert "NorESM1-M" in child_path
    assert "tas" in child_path

    parent = load_scmdataframe(
        glob.glob(os.path.join(test_marble_cmip5_crunch_output, "*", parent_path))[0]
    )

    normalisation_path = res.metadata["(normalisation) netcdf-scm crunched file"]
    assert "piControl" in normalisation_path
    assert "r1i1p1" in normalisation_path
    assert "NorESM1-M" in normalisation_path
    assert "tas" in normalisation_path

    normalisation_path = glob.glob(
        os.path.join(test_marble_cmip5_crunch_output, "*", normalisation_path)
    )[0]

    normalisation = load_scmdataframe(normalisation_path)

    assert parent.metadata["branch_time"] == 255135.0

    ref_values = _get_ref_values(
        res, normalisation, parent, normalisation_path, normalisation_method
    )

    _check_expected_values(res, child, range(2006, 2101), ref_values=ref_values)

    _check_expected_values(res, parent, range(1850, 2006), ref_values=ref_values)


def test_stitching_with_normalisation_in_file_BCC_CSM2_MR(
    tmpdir, caplog, test_cmip6_crunch_output, normalisation_method
):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation_in_file_BCC_CSM2_MR"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*BCC-CSM2-MR.*1pctCO2-bgc.*r1i1p1f1.*tas.*",
                "--out-format",
                "mag-files-average-year-mid-year",
                "--normalise",
                normalisation_method,
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(os.path.join(output_dir, "**", "*.MAG"), recursive=True)
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    _do_generic_normalised_data_tests(res)
    assert res.metadata["normalisation method"] == normalisation_method

    child_path = res.metadata["(child) netcdf-scm crunched file"]
    assert "1pctCO2-bgc" in child_path
    assert "r1i1p1f1" in child_path
    assert "BCC-CSM2-MR" in child_path
    assert "tas" in child_path

    child = load_scmdataframe(
        glob.glob(os.path.join(test_cmip6_crunch_output, "*", child_path))[0]
    )

    normalisation_path = res.metadata["(normalisation) netcdf-scm crunched file"]
    assert "piControl" in normalisation_path
    assert "r1i1p1f1" in normalisation_path
    assert "BCC-CSM2-MR" in normalisation_path
    assert "tas" in normalisation_path

    normalisation_path = glob.glob(
        os.path.join(test_cmip6_crunch_output, "*", normalisation_path)
    )[0]
    normalisation = load_scmdataframe(normalisation_path)

    assert child.metadata["branch_time_in_parent"] == 0.0
    assert child.metadata["parent_time_units"] == "days since 1850-01-01"

    ref_values = _get_ref_values(
        child, normalisation, child, normalisation_path, normalisation_method
    )

    _check_expected_values(
        res, child, range(1850, 2015), ref_values=ref_values, operation="time_mean_ac"
    )

    warn_str = (
        "Assuming BCC metadata is wrong and branch time units are actually years, "
        "not days"
    )
    bcc_warning = [r for r in caplog.record_tuples if r[2] == warn_str]
    # this file has a branch time of zero so the warning shouldn't be zero
    assert not bcc_warning


def test_stitching_with_normalisation_no_picontrol(
    tmpdir, caplog, test_cmip6_crunch_output, normalisation_method
):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation_no_picontrol"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*GFDL-CM4.*1pctCO2.*r1i1p1f1.*",
                "--normalise",
                normalisation_method,
            ],
        )

    assert result.exit_code != 0
    error_msg = re.compile(
        ".*No parent data \\(piControl\\) available for "
        ".*CMIP6/CMIP/NOAA-GFDL/GFDL-CM4/1pctCO2/r1i1p1f1/Amon/tas/gr1/v20180701/netcdf-scm_tas_Amon_GFDL-CM4_1pctCO2_r1i1p1f1_gr1_000101-015012.nc"
        ", we looked in "
        ".*CMIP6/CMIP/NOAA-GFDL/GFDL-CM4/piControl/r1i1p1f1/Amon/tas/gr1/\\*/netcdf-scm_tas_Amon_GFDL-CM4_piControl_r1i1p1f1_gr1_\\*.nc"
    )
    no_parent_error = [r for r in caplog.record_tuples if error_msg.match(r[2])]
    assert len(no_parent_error) == 1
    assert no_parent_error[0][1] == logging.ERROR


def test_stitching_with_normalisation_no_branching_time(
    tmpdir, caplog, test_cmip6_crunch_output, normalisation_method
):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation_no_branching_time"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*CNRM-CM6-1.*hist-aer.*tas.*",
                "--normalise",
                normalisation_method,
            ],
        )

    assert result.exit_code != 0
    error_msg = re.compile(
        ".*Branching time `188301` not available in piControl data in "
        "CMIP6/CMIP/CNRM-CERFACS/CNRM-CM6-1/piControl/r1i1p1f2/Amon/tas/gr/v20180814/netcdf-scm_tas_Amon_CNRM-CM6-1_piControl_r1i1p1f2_gr_230001-231012.nc"
    )
    no_branch_time_error = [r for r in caplog.record_tuples if error_msg.match(r[2])]
    assert len(no_branch_time_error) == 1
    assert no_branch_time_error[0][1] == logging.ERROR


def test_stitching_with_normalisation_not_enough_branching_time(
    tmpdir, caplog, test_cmip6_crunch_output
):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching_with_normalisation_not_enough_branching_time"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*MIROC6.*r1i1p1f1.*rlut.*",
                "--normalise",
                "31-yr-mean-after-branch-time",
            ],
        )

    assert result.exit_code != 0
    error_msg = re.compile(
        ".*Only `320001` to `320212` is available after the branching time `320001` in piControl "
        "data in "
        "CMIP6/CMIP/MIROC/MIROC6/piControl/r1i1p1f1/Amon/rlut/gn/v20181212/netcdf-scm_rlut_Amon_MIROC6_piControl_r1i1p1f1_gn_320001-320212.nc"
    )
    not_enough_norm_data_error = [
        r for r in caplog.record_tuples if error_msg.match(r[2])
    ]
    assert len(not_enough_norm_data_error) == 1
    assert not_enough_norm_data_error[0][1] == logging.ERROR


@pytest.mark.parametrize(
    "out_format",
    (
        "magicc-input-files",
        "magicc-input-files-average-year-start-year",
        "magicc-input-files-average-year-mid-year",
        "magicc-input-files-average-year-end-year",
        "magicc-input-files-point-start-year",
        "magicc-input-files-point-mid-year",
        "magicc-input-files-point-end-year",
        "mag-files",
        "mag-files-average-year-start-year",
        "mag-files-average-year-mid-year",
        "mag-files-average-year-end-year",
        "mag-files-point-start-year",
        "mag-files-point-mid-year",
        "mag-files-point-end-year",
    ),
)
def test_stitching_file_types(
    tmpdir, caplog, test_cmip6_crunch_output, out_format, normalisation_method
):
    if normalisation_method.startswith("21-yr-running-mean"):
        pytest.xfail("Not enough data to take running mean")

    output_dir = str(tmpdir)
    crunch_contact = "test_stitching"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "--out-format",
                out_format,
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*BCC-CSM2-MR.*(ssp126|historical).*tas.*",
                "--normalise",
                normalisation_method,
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(
        os.path.join(
            output_dir, "**", "*.IN" if out_format.startswith("magicc") else "*.MAG"
        ),
        recursive=True,
    )
    assert len(out_files) == 4 if out_format.startswith("magicc") else 2

    for p in out_files:
        res = MAGICCData(p)
        _do_generic_normalised_data_tests(res)

        if "ssp126" in p:
            _do_generic_stitched_data_tests(res)

        if p.endswith(".MAG"):
            _do_filepath_time_consistency_check(res, p, annual="year" in out_format)


@pytest.mark.parametrize(
    "out_format",
    (
        "magicc-input-files",
        "magicc-input-files-average-year-start-year",
        "magicc-input-files-average-year-mid-year",
        "magicc-input-files-average-year-end-year",
        "magicc-input-files-point-start-year",
        "magicc-input-files-point-mid-year",
        "magicc-input-files-point-end-year",
        "mag-files",
        "mag-files-average-year-start-year",
        "mag-files-average-year-mid-year",
        "mag-files-average-year-end-year",
        "mag-files-point-start-year",
        "mag-files-point-mid-year",
        "mag-files-point-end-year",
    ),
)
def test_stitching_file_types_cmip5(
    tmpdir, caplog, test_marble_cmip5_crunch_output, out_format, normalisation_method
):
    output_dir = str(tmpdir)
    crunch_contact = "test_stitching"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_marble_cmip5_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "MarbleCMIP5",
                "--out-format",
                out_format,
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*(hist|rcp45).*tas.*NorESM1.*",
                "--normalise",
                normalisation_method,
            ],
        )

    assert result.exit_code == 0

    out_files = glob.glob(
        os.path.join(
            output_dir, "**", "*.IN" if out_format.startswith("magicc") else "*.MAG"
        ),
        recursive=True,
    )
    assert len(out_files) == 4 if out_format.startswith("magicc") else 2

    for p in out_files:
        res = MAGICCData(p)
        _do_generic_normalised_data_tests(res, cmip5=True)

        if "rcp45" in p:
            _do_generic_stitched_data_tests(res, cmip5=True)

        if p.endswith(".MAG"):
            _do_filepath_time_consistency_check(res, p, annual="year" in out_format)


@pytest.mark.parametrize(
    "out_format",
    (
        "magicc-input-files",
        "magicc-input-files-average-year-start-year",
        "magicc-input-files-average-year-mid-year",
        "magicc-input-files-average-year-end-year",
        "magicc-input-files-point-start-year",
        "magicc-input-files-point-mid-year",
        "magicc-input-files-point-end-year",
        "mag-files",
        "mag-files-average-year-start-year",
        "mag-files-average-year-mid-year",
        "mag-files-average-year-end-year",
        "mag-files-point-start-year",
        "mag-files-point-mid-year",
        "mag-files-point-end-year",
    ),
)
def test_prefix(tmpdir, caplog, test_cmip6_crunch_output, out_format):
    output_dir = str(tmpdir)
    crunch_contact = "test_prefix"
    prefix = "NORMED"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*BCC-CSM2-MR.*ssp126.*tas.*",
                "--out-format",
                out_format,
                "--prefix",
                prefix,
            ],
        )

    assert result.exit_code == 0, result.stderr

    out_files = glob.glob(
        os.path.join(
            output_dir, "**", "*.IN" if out_format.startswith("magicc") else "*.MAG"
        ),
        recursive=True,
    )
    assert len(out_files) == 2 if out_format.startswith("magicc") else 1
    assert [os.path.basename(p).startswith("{}_".format(prefix)) for p in out_files]


def test_target_units(tmpdir, caplog, test_cmip6_crunch_output):
    target_unit = "J / yr / m^2"
    target_units = pd.DataFrame(
        [["fgco2", "g / m**2 / s"], ["hfds", target_unit]], columns=["variable", "unit"]
    )
    target_units_csv = os.path.join(tmpdir, "target_units.csv")
    target_units.to_csv(target_units_csv, index=False)

    output_dir = str(tmpdir)
    crunch_contact = "test_target_units"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level(logging.WARNING):
        result = runner.invoke(
            stitch_netcdf_scm_ncs,
            [
                test_cmip6_crunch_output,
                output_dir,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--number-workers",
                1,
                "--regexp",
                ".*EC-Earth3-Veg.*ssp585.*r1i1p1f1.*hfds.*",
                "--target-units-specs",
                target_units_csv,
            ],
        )

    assert result.exit_code == 0, result.stderr

    out_files = glob.glob(os.path.join(output_dir, "**", "*.MAG"), recursive=True)
    assert len(out_files) == 1

    res = MAGICCData(out_files[0])

    _do_generic_stitched_data_tests(res)
    _do_filepath_time_consistency_check(res, out_files[0])
    assert res["unit"].unique().tolist() == ["J / yr / m^2"]
    assert "Converting units" in result.stderr


@pytest.fixture(scope="module")
def picontrol_data(test_cmip6_crunch_output):
    test_file = os.path.join(
        test_cmip6_crunch_output,
        "Amon",
        "CMIP6",
        "CMIP",
        "NCAR",
        "CESM2",
        "piControl",
        "r1i1p1f1",
        "Amon",
        "tas",
        "gn",
        "v20190320",
        "netcdf-scm_tas_Amon_CESM2_piControl_r1i1p1f1_gn_080001-099912.nc",
    )

    loaded = load_scmdataframe(test_file)
    loaded.metadata["netcdf-scm crunched file"] = test_file

    return loaded


@pytest.fixture(scope="module")
def historical_data(test_cmip6_crunch_output):
    test_file = os.path.join(
        test_cmip6_crunch_output,
        "Amon",
        "CMIP6",
        "CMIP",
        "NCAR",
        "CESM2",
        "historical",
        "r10i1p1f1",
        "Amon",
        "tas",
        "gn",
        "v20190313",
        "netcdf-scm_tas_Amon_CESM2_historical_r10i1p1f1_gn_185001-201412.nc",
    )

    loaded = load_scmdataframe(test_file)
    loaded.metadata["netcdf-scm crunched file"] = test_file

    return loaded


@pytest.mark.parametrize("expected_time,parent", ((306600.0, True), (674885.0, False),))
def test_get_branch_time_cmip6(historical_data, expected_time, parent):
    raw = netCDF4.num2date(expected_time, "days since 0001-01-01 00:00:00", "365_day",)
    expected = dt.datetime(raw.year, raw.month, raw.day)

    if parent:
        res = get_branch_time(historical_data, parent)
    else:
        res = get_branch_time(
            historical_data,
            parent,
            source_path=historical_data.metadata["netcdf-scm crunched file"],
        )

    assert res == expected


def test_get_branch_time_bcc_warning(caplog, test_cmip6_crunch_output):
    bcc_dat = load_scmdataframe(
        os.path.join(
            test_cmip6_crunch_output,
            "Amon",
            "CMIP6",
            "CMIP",
            "BCC",
            "BCC-CSM2-MR",
            "historical",
            "r1i1p1f1",
            "Amon",
            "tas",
            "gn",
            "v20181126",
            "netcdf-scm_tas_Amon_BCC-CSM2-MR_historical_r1i1p1f1_gn_185001-201412.nc",
        )
    )
    expected = dt.datetime(int(bcc_dat.metadata["branch_time_in_parent"]), 1, 1)

    with caplog.at_level(logging.WARNING):
        res = get_branch_time(bcc_dat)

    assert res == expected
    warn_str = (
        "Assuming BCC metadata is wrong and branch time units are actually years, "
        "not days"
    )
    bcc_warning = [r for r in caplog.record_tuples if r[2] == warn_str]
    assert len(bcc_warning) == 1
    assert bcc_warning[0][1] == logging.WARNING


def test_get_branch_time_cmip5(test_marble_cmip5_crunch_output):
    cmip5_dat = load_scmdataframe(
        os.path.join(
            test_marble_cmip5_crunch_output,
            "Amon",
            "cmip5",
            "historical",
            "Amon",
            "tas",
            "NorESM1-M",
            "r1i1p1",
            "netcdf-scm_tas_Amon_NorESM1-M_historical_r1i1p1_185001-200512.nc",
        )
    )

    parent_path = os.path.join(
        test_marble_cmip5_crunch_output,
        "Amon",
        "cmip5",
        "piControl",
        "Amon",
        "tas",
        "NorESM1-M",
        "r1i1p1",
        "netcdf-scm_tas_Amon_NorESM1-M_piControl_r1i1p1_070001-120012.nc",
    )

    raw = netCDF4.num2date(255135.0, "days since 1-01-01 00:00:00", "365_day",)
    expected = dt.datetime(raw.year, raw.month, raw.day)

    res = get_branch_time(cmip5_dat, parent_path=parent_path)

    assert res == expected


def test_get_parent_file_path(historical_data, picontrol_data):
    parent_replacements = get_parent_replacements(historical_data)
    res = get_parent_file_path(
        historical_data.metadata["netcdf-scm crunched file"],
        parent_replacements,
        "CMIP6Output",
    )

    assert res == picontrol_data.metadata["netcdf-scm crunched file"]


def test_get_parent_file_path_no_file_error(test_cmip6_crunch_output):
    # we don't have the parent data for this file in the repo
    test_path = os.path.join(
        test_cmip6_crunch_output,
        "Omon",
        "CMIP6",
        "CMIP",
        "NCAR",
        "CESM2",
        "historical",
        "r10i1p1f1",
        "Omon",
        "tos",
        "gn",
        "v20190313",
        "netcdf-scm_tos_Omon_CESM2_historical_r10i1p1f1_gn_199801-200112.nc",
    )

    start = load_scmdataframe(test_path)
    parent_replacements = get_parent_replacements(start)

    error_msg = re.escape(
        "No parent data ({}) available for {}, we looked in {}".format(
            parent_replacements["parent_experiment_id"],
            test_path,
            test_path.replace("v20190313", "*")
            .replace("199801-200112", "*")
            .replace("historical", "piControl")
            .replace("r10i1p1f1", "r1i1p1f1"),
        )
    )
    with pytest.raises(IOError, match=error_msg):
        get_parent_file_path(
            test_path, parent_replacements, "CMIP6Output",
        )


@pytest.mark.parametrize("return_picontrol_info", (False, True))
def test_get_continuous_timeseries_with_meta(
    test_cmip6_crunch_output, return_picontrol_info
):
    start = os.path.join(
        test_cmip6_crunch_output,
        "Amon",
        "CMIP6",
        "ScenarioMIP",
        "BCC",
        "BCC-CSM2-MR",
        "ssp126",
        "r1i1p1f1",
        "Amon",
        "tas",
        "gn",
        "v20190314",
        "netcdf-scm_tas_Amon_BCC-CSM2-MR_ssp126_r1i1p1f1_gn_201501-210012.nc",
    )

    res, picontrol_branch_time, picontrol_file = get_continuous_timeseries_with_meta(
        start, "CMIP6Output", return_picontrol_info=return_picontrol_info,
    )

    assert res["year"].min() == 1850
    assert res["year"].max() == 2100
    assert res.get_unique_meta("scenario", no_duplicates=True) == "ssp126"

    if return_picontrol_info:
        assert picontrol_file == os.path.join(
            test_cmip6_crunch_output,
            "Amon",
            "CMIP6",
            "CMIP",
            "BCC",
            "BCC-CSM2-MR",
            "piControl",
            "r1i1p1f1",
            "Amon",
            "tas",
            "gn",
            "v20181016",
            "netcdf-scm_tas_Amon_BCC-CSM2-MR_piControl_r1i1p1f1_gn_185001-244912.nc",
        )
        assert picontrol_branch_time == dt.datetime(2289, 1, 1)
    else:
        assert picontrol_file is None
        assert picontrol_branch_time is None


@pytest.mark.parametrize("return_picontrol_info", (False, True))
def test_get_continuous_timeseries_with_meta_picontrol(
    test_cmip6_crunch_output, assert_scmdata_frames_allclose, return_picontrol_info
):
    # no file and branching returned irrespective of `return_picontrol_info` value
    start = os.path.join(
        test_cmip6_crunch_output,
        "Amon/CMIP6/CMIP/BCC/BCC-CSM2-MR/piControl/r1i1p1f1/Amon/tas/gn/v20181016/netcdf-scm_tas_Amon_BCC-CSM2-MR_piControl_r1i1p1f1_gn_185001-244912.nc",
    )

    res, picontrol_branch_time, picontrol_file = get_continuous_timeseries_with_meta(
        start, "CMIP6Output", return_picontrol_info
    )

    expected_res = load_scmdataframe(start)
    expected_res.metadata["netcdf-scm crunched file"] = start.replace(
        "{}".format(os.path.join(test_cmip6_crunch_output, "Amon/")), ""
    )

    assert_scmdata_frames_allclose(res, expected_res)
    assert res["year"].min() == 1850
    assert res["year"].max() == 2449
    assert res.get_unique_meta("scenario", no_duplicates=True) == "piControl"

    assert picontrol_file is None
    assert picontrol_branch_time is None


def test_issue_48_get_continuous_timeseries_with_expanded(
    test_data_root_dir, assert_scmdata_frames_allclose
):
    data_root_dir = os.path.join(test_data_root_dir, "48-whitespace-in-id-test-data")
    start = os.path.join(
        data_root_dir,
        "CMIP6",
        "CMIP",
        "CNRM-CERFACS",
        "CNRM-CM6-1-HR",
        "1pctCO2",
        "r1i1p1f2",
        "Amon",
        "rsut",
        "gr",
        "v20191021",
        "netcdf-scm_rsut_Amon_CNRM-CM6-1-HR_1pctCO2_r1i1p1f2_gr_185001-199912.nc",
    )

    res, picontrol_branch_time, picontrol_file = get_continuous_timeseries_with_meta(
        start, "CMIP6Output", True
    )

    expected_res = load_scmdataframe(start)
    expected_res.metadata["netcdf-scm crunched file"] = start.replace(
        "{}".format(data_root_dir + "/"), ""
    )

    assert_scmdata_frames_allclose(res, expected_res)
    assert res["year"].min() == 1850
    assert res["year"].max() == 1999
    assert res.get_unique_meta("scenario", no_duplicates=True) == "1pctCO2"

    assert "netcdf-scm_rsut_Amon_CNRM-CM6-1-HR_piControl_r1i1p1f2_gr" in picontrol_file
    assert picontrol_branch_time.year == 1850



from os.path import join

import pytest
from click.testing import CliRunner

from netcdf_scm.cli import stitch_netcdf_scm_ncs


@pytest.mark.parametrize(
    "out_format",
    (
        [
            "mag-files",
            "magicc-input-files-average-year-start-year",
            "magicc-input-files-point-start-year",
        ]
    ),
)
def test_stitching_and_normalise_results(
    tmpdir,
    update_expected_files,
    test_marble_cmip5_crunch_output,
    test_marble_cmip5_stitch_output,
    run_wrangling_comparison,
    out_format,
    normalisation_method,
):
    input_dir = test_marble_cmip5_crunch_output
    output_dir = join(str(tmpdir), normalisation_method)

    runner = CliRunner()
    result = runner.invoke(
        stitch_netcdf_scm_ncs,
        [
            input_dir,
            output_dir,
            "cmip6output stitching with {} normalisation regression test".format(
                normalisation_method
            ),
            "--drs",
            "MarbleCMIP5",
            "-f",
            "--number-workers",
            1,
            "--out-format",
            out_format,
            # have to avoid files which will fail normalising
            "--regexp",
            ".*rcp45.*tas.*NorESM1.*",
            "--normalise",
            normalisation_method,
        ],
    )
    assert result.exit_code == 0, result.output

    run_wrangling_comparison(
        join(output_dir, "cmip5"),
        join(
            test_marble_cmip5_stitch_output,
            "normalised",
            normalisation_method,
            out_format,
            "cmip5",
        ),
        update=update_expected_files,
    )



"""Functions used to 'wrangle' netCDF-SCM netCDF files into other formats"""
import logging
from os.path import isfile, join

import numpy as np
from scmdata import ScmDataFrame

from . import mat4py

logger = logging.getLogger(__name__)


def convert_tuningstruc_to_scmdf(  # pylint:disable=too-many-arguments,too-many-locals
    filepath, variable=None, region=None, unit=None, scenario=None, model=None
):
    """
    Convert a matlab tuningstruc to an ScmDataFrame

    Parameters
    ----------
    filepath : str
        Filepath from which to load the data

    variable : str
        Name of the variable contained in the tuningstruc. If None,
        `convert_tuningstruc_to_scmdf` will attempt to determine it from the input
        file.

    region : str
        Region to which the data in the tuningstruc applies. If None,
        `convert_tuningstruc_to_scmdf` will attempt to determine it from the input
        file.

    unit : str
        Units of the data in the tuningstruc. If None,
        `convert_tuningstruc_to_scmdf` will attempt to determine it from the input
        file.

    scenario : str
        Scenario to which the data in the tuningstruc applies. If None,
        `convert_tuningstruc_to_scmdf` will attempt to determine it from the input
        file.

    model : str
        The (integrated assessment) model which generated the emissions scenario
        associated with the data in the tuningstruc. If None,
        `convert_tuningstruc_to_scmdf` will attempt to determine it from the input
        file and if it cannot, it will be set to "unspecified".

    Raises
    ------
    KeyError
        If a metadata variable is not supplied and it cannot be determined from the
        tuningstruc.

    Returns
    -------
    :obj:`ScmDataFrame`
        ScmDataFrame with the tuningstruc data
    """
    dataset = mat4py.loadmat(filepath)

    for m, climate_model in enumerate(dataset["tuningdata"]["modelcodes"]):
        metadata = {
            "variable": [variable],
            "region": [region],
            "unit": [unit],
            "climate_model": [climate_model],
            "scenario": [scenario],
            "model": [model],
        }
        for k, v in metadata.items():
            if v == [None]:
                try:
                    metadata[k] = [dataset["tuningdata"]["model"][m][k]]
                except KeyError:
                    if k == "model":
                        metadata[k] = ["unspecified"]
                        continue

                    error_msg = "Cannot determine {} from file: " "{}".format(
                        k, filepath
                    )
                    raise KeyError(error_msg)

        data = np.asarray(dataset["tuningdata"]["model"][m]["data"])
        if len(data) != 2:
            data = data.T
        scmdf = ScmDataFrame(data=data[1], index=data[0], columns=metadata)

        try:
            ref_df.append(scmdf, inplace=True)
        except NameError:
            ref_df = scmdf
    return ref_df


def convert_scmdf_to_tuningstruc(scmdf, outdir, prefix=None, force=False):
    """
    Convert an ScmDataFrame to a matlab tuningstruc

    One tuningstruc file will be created for each unique
    ["model", "scenario", "variable", "region", "unit"] combination in the input
    ScmDataFrame.

    Parameters
    ----------
    scmdf : :obj:`ScmDataFrame`
        ScmDataFrame to convert to a tuningstruc

    outdir : str
        Directory in which to save the tuningstruc

    prefix : str
        Prefix for the filename. The rest of the filename is generated from the
        metadata. `.mat` is also appended automatically. If ``None``, no prefix
        is used.

    force : bool
        If True, overwrite any existing files

    Returns
    -------
    list
        List of files which were not re-written as they already exist

    Raises
    ------
    AssertionError
        If timeseries are not unique for a given ["climate_model", "model",
        "scenario", "variable", "region", "unit"] combination.
    """
    already_written = []

    group_cols = ["model", "scenario", "variable", "region", "unit"]
    for label, df in scmdf.timeseries().groupby(group_cols):
        ids = dict(zip(group_cols, label))

        dataset = {}
        dataset["tuningdata"] = {}
        dataset["tuningdata"]["modelcodes"] = []
        dataset["tuningdata"]["model"] = []

        for m, (climate_model, cmdf) in enumerate(df.groupby("climate_model")):
            # impossible to make dataframe with duplicate rows, this is just in
            # case
            if cmdf.shape[0] != 1:  # pragma: no cover # emergency valve
                raise AssertionError(
                    "Should only have a single unique timeseries for a given "
                    '["climate_model", "model", "scenario", "variable", '
                    '"region", "unit"] combination'
                )

            dataset["tuningdata"]["modelcodes"].append(climate_model)
            dataset["tuningdata"]["model"].append({})

            dataset["tuningdata"]["model"][m]["model"] = ids["model"]
            dataset["tuningdata"]["model"][m]["scenario"] = ids["scenario"]
            dataset["tuningdata"]["model"][m]["variable"] = ids["variable"]
            dataset["tuningdata"]["model"][m]["region"] = ids["region"]
            dataset["tuningdata"]["model"][m]["unit"] = ids["unit"]

            dataset["tuningdata"]["model"][m]["notes"] = (
                "{} {} {} {} ({}) tuningstruc (written with scmcallib)"
                "".format(
                    ids["scenario"],
                    ids["model"],
                    ids["region"],
                    ids["variable"],
                    ids["unit"],
                )
            )
            dataset["tuningdata"]["model"][m]["data"] = [
                list(e)
                for e in zip(
                    [float(t.year) for t in cmdf.columns], list(cmdf.values.squeeze()),
                )
            ]
            dataset["tuningdata"]["model"][m]["col_code"] = ["YEARS", ids["variable"]]

        outfile = get_tuningstruc_name_from_df(df, outdir, prefix)

        if isfile(outfile) and not force:
            logger.info("Skipped (already exists, not overwriting) %s", outfile)
            already_written.append(outfile)
        else:
            logger.info("Writing %s", outfile)
            mat4py.savemat(outfile, dataset)

    return already_written


def get_tuningstruc_name_from_df(df, outdir, prefix):
    """
    Get the name of a tuningstruc from a ``pd.DataFrame``

    Parameters
    ----------
    df : :obj:`pd.DataFrame`
        *pandas* DataFrame to convert to a tuningstruc

    outdir : str
        Base path on which to append the metadata and `.mat`.

    prefix : str
        Prefix to prepend to the name. If ``None``, no prefix is prepended.

    Returns
    -------
    str
        tuningstruc name

    Raises
    ------
    ValueError
        A name cannot be determined because e.g. more than one scenario is contained
        in the dataframe
    """
    prefix = "{}_".format(prefix) if prefix is not None else ""

    def _get_col(col):
        try:
            vals = df[col].unique()
        except KeyError:
            vals = df.index.get_level_values(col).unique()
        if len(vals) != 1:
            raise ValueError("More than one {} in df".format(col))

        return vals[0]

    scenario = _get_col("scenario")
    try:
        member_id = _get_col("member_id")
    except KeyError:
        member_id = "member-id"
    variable = _get_col("variable")
    region = _get_col("region")
    raw_name = (
        (
            "{}_{}_{}_{}".format(variable, scenario, member_id, region)
            .replace(" ", "_")
            .replace("|", "_")
        )
    ).upper() + ".mat"

    return join(outdir, "{}{}".format(prefix, raw_name))



import datetime as dt
import re
from abc import ABC, abstractmethod
from unittest.mock import patch

import pandas as pd
import pytest

from netcdf_scm.stitching import get_branch_time


@pytest.mark.usefixtures("junk_data_normalisation", "junk_data_picontrol_normalisation")
class NormaliserTester(ABC):
    tclass = None

    @abstractmethod
    def test_method_name(self):
        """
        assert self.tclass().method_name == expected_name
        """

    def test_get_reference_values_no_branch_time_data(
        self, junk_data_normalisation, junk_data_picontrol_normalisation
    ):
        picontrol_branch_time = get_branch_time(junk_data_normalisation)
        pi_data_to_use = junk_data_picontrol_normalisation.filter(
            year=picontrol_branch_time.year, keep=False
        )

        error_msg = re.compile(
            ".*Branching time `185001` not available in piControl data in "
            "{}".format(pi_data_to_use.metadata["netcdf-scm crunched file"])
        )

        with pytest.raises(ValueError, match=error_msg):
            self.tclass().get_reference_values(
                junk_data_normalisation, pi_data_to_use, picontrol_branch_time,
            )

    def test_normalise_against_picontrol(
        self, junk_data_normalisation, junk_data_picontrol_normalisation
    ):
        mock_ref_values = junk_data_normalisation.timeseries()
        mock_ref_values.iloc[:, :] = 1

        picontrol_branch_time = dt.datetime(2, 2, 2)  # mocked so irrelevant

        with patch.object(
            self.tclass, "get_reference_values", return_value=mock_ref_values
        ) as mock_get_reference_values:

            res = self.tclass().normalise_against_picontrol(
                junk_data_normalisation,
                junk_data_picontrol_normalisation,
                picontrol_branch_time,
            )

        mock_get_reference_values.assert_called_with(
            junk_data_normalisation,
            junk_data_picontrol_normalisation,
            picontrol_branch_time,
        )

        expected_metadata = {
            **{
                "(child) {}".format(k): v
                for k, v in junk_data_normalisation.metadata.items()
            },
            **{
                "(normalisation) {}".format(k): v
                for k, v in junk_data_picontrol_normalisation.metadata.items()
            },
        }
        expected_metadata["normalisation method"] = self.tclass().method_name

        assert res.metadata == expected_metadata

        # thanks to mocking, everything simply shifted down by 1
        pd.testing.assert_frame_equal(
            res.timeseries(), junk_data_normalisation.timeseries() - 1
        )

    @pytest.mark.parametrize(
        "picontrol_expt_name,error", (("esm-piControl", False), ("piC", True),)
    )
    def test_normalise_against_picontrol_name_error(
        self,
        picontrol_expt_name,
        error,
        junk_data_picontrol_normalisation,
        junk_data_normalisation,
    ):
        picontrol_branch_time = get_branch_time(junk_data_normalisation)
        tweaked_picontrol = junk_data_picontrol_normalisation.copy()
        tweaked_picontrol.metadata["experiment_id"] = picontrol_expt_name

        tweaked_picontrol = tweaked_picontrol.resample("MS")
        junk_data_normalisation = junk_data_normalisation.resample("MS")

        def call():
            self.tclass().normalise_against_picontrol(
                junk_data_normalisation, tweaked_picontrol, picontrol_branch_time,
            )

        if error:
            error_msg = re.escape(
                "If you would like to normalise against an experiment other than piControl, please raise an "
                "issue at https://gitlab.com/netcdf-scm/netcdf-scm/-/issues"
            )
            with pytest.raises(NotImplementedError, match=error_msg):
                call()

        else:
            # no error
            call()



import pytest

from netcdf_scm.normalisation import NormaliserRunningMean, get_normaliser


@pytest.mark.parametrize(
    "key,expected_cls", (("21-yr-running-mean", NormaliserRunningMean),)
)
def test_get_normaliser(key, expected_cls):
    assert isinstance(get_normaliser(key), expected_cls)


def test_get_normaliser_error():
    error_msg = "Unrecognised key: 'junk'"
    with pytest.raises(ValueError, match=error_msg):
        get_normaliser("junk")



import json
import logging
import os.path
import re
from glob import glob
from unittest.mock import patch

import pandas as pd
import pytest

from netcdf_scm.iris_cube_wrappers import CMIP6OutputCube
from netcdf_scm.retractions import (
    _create_search_url,
    _perform_esgf_queries,
    check_depends_on_retracted,
    check_retracted_files,
    check_retractions,
)


def mock_resp(instances, retracted):
    docs = [{"instance_id": i, "retracted": r} for i, r in zip(instances, retracted)]
    return {"response": {"docs": docs}}, instances


@pytest.fixture()
def esgf_result(test_data_root_dir):
    # JSON output from an esgf query
    with open(os.path.join(test_data_root_dir, "retractions.json")) as fh:
        return json.load(fh)


@patch("netcdf_scm.retractions._perform_esgf_queries")
def test_check_retractions(mock_query):
    instance_ids = [
        "CMIP6.DAMIP.NASA-GISS.GISS-E2-1-G.hist.r1i1p1f1.Amon.tas.gn.v20191226",
        "CMIP6.DAMIP.NASA-GISS.GISS-E2-1-G.hist.r1i1p1f1.Amon.rsut.gn.v20191226",
    ]
    mock_query.return_value = [mock_resp(instance_ids, [False, True])]
    res = check_retractions(instance_ids, esgf_query_batch_size=50, nworkers=1)

    expected = [
        "CMIP6.DAMIP.NASA-GISS.GISS-E2-1-G.hist.r1i1p1f1.Amon.rsut.gn.v20191226"
    ]
    assert res == expected

    mock_query.assert_called_with(instance_ids, 50, 1)


@patch("netcdf_scm.retractions._perform_esgf_queries")
def test_check_retractions_couldnt_find(mock_query, caplog):
    instance_ids = [
        "CMIP6.DAMIP.NASA-GISS.GISS-E2-1-G.hist.r1i1p1f1.Amon.tas.gn.v20191226",
        "CMIP6.DAMIP.NASA-GISS.GISS-E2-1-G.hist.r1i1p1f1.Amon.rsut.gn.v20191226",
    ]

    mock_return_value = [mock_resp(instance_ids, [False, True])]
    # drop out one of the instance ids, as if the search didn't find results
    mock_return_value[0][0]["response"]["docs"] = mock_return_value[0][0]["response"][
        "docs"
    ][1:]
    mock_query.return_value = mock_return_value
    res = check_retractions(instance_ids, esgf_query_batch_size=50, nworkers=1)

    assert len(caplog.record_tuples) == 1
    name, level, msg = caplog.record_tuples[0]
    assert name == "netcdf_scm.retractions"
    assert level == logging.WARNING
    assert msg.startswith(
        "Couldn't find "
        "CMIP6.DAMIP.NASA-GISS.GISS-E2-1-G.hist.r1i1p1f1.Amon.tas.gn.v20191226 "
        "(search url: "
    )

    expected = [
        "CMIP6.DAMIP.NASA-GISS.GISS-E2-1-G.hist.r1i1p1f1.Amon.rsut.gn.v20191226"
    ]
    assert res == expected

    mock_query.assert_called_with(instance_ids, 50, 1)


@pytest.mark.parametrize(
    "search_kwargs,exp",
    (
        (
            {},
            "https://esgf.nci.org.au/esg-search/search?format=application%2Fsolr%2Bjson&offset=0&limit=100&type=Dataset&replica=False",
        ),
        (
            {"experiment_id": "ssp585"},
            "https://esgf.nci.org.au/esg-search/search?experiment_id=ssp585&format=application%2Fsolr%2Bjson&offset=0&limit=100&type=Dataset&replica=False",
        ),
        (
            {"experiment_id": "ssp585", "grid_label": "gn"},
            "https://esgf.nci.org.au/esg-search/search?experiment_id=ssp585&grid_label=gn&format=application%2Fsolr%2Bjson&offset=0&limit=100&type=Dataset&replica=False",
        ),
    ),
)
def test_search_url(search_kwargs, exp):
    res = _create_search_url(**search_kwargs)

    assert res == exp


@pytest.mark.parametrize("batch_size", [5, 10])
@patch("netcdf_scm.retractions._query_esgf")
def test_retractions_batch_size(mock_query, batch_size):
    mock_query.side_effect = lambda ids: ({}, ids)
    instance_ids = [
        "CMIP6.DAMIP.NASA-GISS.GISS-E2-1-G.hist.r{}i1p1f1.Amon.tas.gn.v20191226".format(
            i
        )
        for i in range(30)
    ]

    res = _perform_esgf_queries(
        instance_ids, esgf_query_batch_size=batch_size, nworkers=1
    )

    assert len(res) == 30 // batch_size

    for i, r in enumerate(res):
        assert r[0] == {}
        assert len(r[1]) == batch_size


@patch("netcdf_scm.retractions._query_esgf")
def test_retractions_integration(mock_query, esgf_result):
    mock_query.side_effect = lambda ids: (esgf_result, ids)
    instance_ids = [
        "CMIP6.DAMIP.NASA-GISS.GISS-E2-1-G.hist.r1i1p1f1.Amon.tas.gn.v20191226",
        "CMIP6.DAMIP.NASA-GISS.GISS-E2-1-G.hist.r1i1p1f1.Amon.rsut.gn.v20191226",
    ]

    res = check_retractions(instance_ids, esgf_query_batch_size=50, nworkers=1)
    expected = [
        "CMIP6.DAMIP.NASA-GISS.GISS-E2-1-G.hist-aer.r2i1p1f2.Amon.rsut.gn.v20191226"
    ]
    assert res == expected

    mock_query.assert_called_with(instance_ids)


list_of_files_mark = pytest.mark.parametrize("list_of_files", [True, False])


@list_of_files_mark
@patch("netcdf_scm.retractions.check_retractions")
def test_check_retracted_files_mag(
    mock_check_retractions, test_cmip6_stitch_output, list_of_files
):
    _check_retracted_files(
        mock_check_retractions,
        test_cmip6_stitch_output,
        "*.MAG",
        list_of_files,
        expected_retracted=3,
    )


@list_of_files_mark
@patch("netcdf_scm.retractions.check_retractions")
def test_check_retracted_files_nc(
    mock_check_retractions, test_data_cmip6output_dir, list_of_files
):
    _check_retracted_files(
        mock_check_retractions,
        test_data_cmip6output_dir,
        "*.nc",
        list_of_files,
        expected_retracted=1,
    )


def _check_retracted_files(
    mock_check_retractions, test_dir, filter_txt, list_of_files, expected_retracted
):
    fnames = glob(os.path.join(test_dir, "**", filter_txt), recursive=True)
    instance_ids = list(set([CMIP6OutputCube.get_instance_id(f) for f in fnames]))
    mock_check_retractions.return_value = [
        "CMIP6.CMIP.BCC.BCC-CSM2-MR.historical.r1i1p1f1.Amon.tas.gn.v20181126"
    ]

    if list_of_files:
        fnames_or_dir = fnames
    else:
        fnames_or_dir = test_dir

    res = check_retracted_files(
        fnames_or_dir, filename_filter=filter_txt, esgf_query_batch_size=50, nworkers=1
    )

    mock_check_retractions.assert_called()
    _, call_args, call_kwargs = mock_check_retractions.mock_calls[0]
    assert call_kwargs == {"esgf_query_batch_size": 50, "nworkers": 1}
    assert sorted(call_args[0]) == sorted(instance_ids)

    assert len(res) == expected_retracted
    for f in res:
        assert "tas_Amon_BCC-CSM2-MR_historical_r1i1p1f1_gn" in f


def _check_depends_on_retracted_files_result(res):
    retracted_mag = res.loc[res["dependency_retracted"], "mag_file"].unique().tolist()
    assert len(retracted_mag) == 6

    def check_retracted_file(f):
        return (
            "tas_Amon_BCC-CSM2-MR_historical_r1i1p1f1_gn" in f
            or "tas_Amon_BCC-CSM2-MR_ssp126_r1i1p1f1_gn" in f
        )

    for f in retracted_mag:
        assert check_retracted_file(f) and f.endswith(".MAG")

    retracted_nc = (
        res.loc[res["dependency_retracted"], "dependency_file"].unique().tolist()
    )
    for f in retracted_nc:
        assert check_retracted_file(f) and f.endswith(".nc")

    assert (
        res["dependency_instance_id"]
        == res["dependency_file"].apply(CMIP6OutputCube.get_instance_id)
    ).all()


@patch("netcdf_scm.retractions.check_retractions")
def test_check_depends_on_retracted(mock_check_retractions, test_cmip6_stitch_output):
    fnames = glob(os.path.join(test_cmip6_stitch_output, "**", "*.MAG"), recursive=True)

    mock_check_retractions.return_value = [
        "CMIP6.CMIP.BCC.BCC-CSM2-MR.historical.r1i1p1f1.Amon.tas.gn.v20181126"
    ]

    res = check_depends_on_retracted(fnames, esgf_query_batch_size=50, nworkers=1)

    assert isinstance(res, pd.DataFrame)
    assert res.columns.tolist() == [
        "mag_file",
        "dependency_file",
        "dependency_instance_id",
        "dependency_retracted",
    ]

    mock_check_retractions.assert_called()
    _, _, call_kwargs = mock_check_retractions.mock_calls[0]
    assert call_kwargs == {"esgf_query_batch_size": 50, "nworkers": 1}

    _check_depends_on_retracted_files_result(res)


@patch("netcdf_scm.retractions.check_retractions")
@pytest.mark.parametrize("raise_on_mismatch", (True, False))
def test_check_depends_on_retracted_not_cmip6(
    mock_check_retractions,
    raise_on_mismatch,
    caplog,
    test_marble_cmip5_stitch_output,
    test_cmip6_stitch_output,
):
    fnames_cmip5 = glob(
        os.path.join(test_marble_cmip5_stitch_output, "**", "*.MAG"), recursive=True
    )
    fnames_cmip6 = glob(
        os.path.join(test_cmip6_stitch_output, "**", "*.MAG"), recursive=True
    )

    def call():
        return check_depends_on_retracted(
            fnames_cmip5 + fnames_cmip6, raise_on_mismatch=raise_on_mismatch
        )

    if raise_on_mismatch:
        error_msg = re.escape(
            "Source data is not CMIP6 for: {}".format(set(fnames_cmip5))
        )
        with pytest.raises(ValueError, match=error_msg):
            call()
    else:
        mock_check_retractions.return_value = [
            "CMIP6.CMIP.BCC.BCC-CSM2-MR.historical.r1i1p1f1.Amon.tas.gn.v20181126"
        ]

        res = call()
        _check_depends_on_retracted_files_result(res)

        assert len(caplog.record_tuples) == 1
        name, level, msg = caplog.record_tuples[0]
        assert name == "netcdf_scm.retractions"
        assert level == logging.ERROR
        assert msg.startswith("Source data is not CMIP6 for: ")

        mock_check_retractions.assert_called()


@patch("netcdf_scm.retractions.check_retractions")
@patch("netcdf_scm.retractions.pymagicc.io.read_mag_file_metadata")
@pytest.mark.parametrize("raise_on_mismatch", (True, False))
def test_check_depends_on_retracted_no_metadata(
    mock_read_mag_file_metadata, mock_check_retractions, raise_on_mismatch, caplog
):
    tfpath = "/CMIP6/C4MIP/BCC/BCC-CSM2-MR/1pctCO2-bgc/r1i1p1f1/Amon/tas/gn/v20190321/tas_Amon_BCC-CSM2-MR_1pctCO2-bgc_r1i1p1f1_gn_185001-200012.nc"

    def get_metadata(fname):
        if fname.startswith("c"):
            return {"(child) crunch_source_files": ("Files: " "['{}']".format(tfpath))}

        return {}

    mock_read_mag_file_metadata.side_effect = get_metadata

    fnames = ["a.MAG", "b.MAG", "c.MAG"]
    fnames_dud = fnames[:2]
    error_msg = re.escape(
        "Could not determine source data for: {}".format(set(fnames_dud))
    )

    def call():
        check_depends_on_retracted(fnames, raise_on_mismatch=raise_on_mismatch)

    if raise_on_mismatch:
        with pytest.raises(ValueError, match=error_msg):
            call()
    else:
        call()
        assert len(caplog.record_tuples) == 1
        name, level, msg = caplog.record_tuples[0]
        assert name == "netcdf_scm.retractions"
        assert level == logging.ERROR
        assert msg.startswith("Could not determine source data for: ")

        mock_check_retractions.assert_called()
        mock_check_retractions.assert_called_with(
            [CMIP6OutputCube.get_instance_id(tfpath)]
        )



"""Miscellaneous readers for files which can't otherwise be read"""
from scmdata import ScmDataFrame

from .iris_cube_wrappers import ScmCube
from .utils import _check_cube_and_adjust_if_needed

try:
    import iris

except ModuleNotFoundError:  # pragma: no cover # emergency valve
    from .errors import raise_no_iris_warning

    raise_no_iris_warning()


def read_cmip6_concs_gmnhsh(  # pylint:disable=too-many-locals
    filepath, region_coord_name="sector"
):
    """
    Read CMIP6 concentrations global and hemispheric mean data

    Parameters
    ----------
    filepath : str
        Filepath from which to read the data

    region_coord_name : str
        The name of the co-ordinate which represents the region in the datafile.

    Returns
    -------
    :obj:`ScmDataFrame`
        :obj:`ScmDataFrame` containing the global and hemispheric mean data

    Raises
    ------
    AssertionError
        Defensive assertion: the code is being used in an unexpected way
    """
    loaded_cube = iris.load_cube(filepath)
    checked_cube = _check_cube_and_adjust_if_needed(loaded_cube)

    region_map = {
        "GM": "World",
        "NH": "World|Northern Hemisphere",
        "SH": "World|Southern Hemisphere",
    }
    unit_map = {"1.e^-6": "ppm", "1.e^-9": "ppb", "1.e^-12": "ppt"}

    timeseries_cubes = {}
    for region_coord in checked_cube.coord(region_coord_name):
        if len(list(region_coord.cells())) != 1:  # pragma: no cover
            raise AssertionError("Should only have one point now")

        original_names = {
            int(v.split(":")[0].strip()): v.split(":")[1].strip()
            for v in region_coord.attributes["original_names"].split(";")
        }
        original_regions = {k: v.split("_")[-1] for k, v in original_names.items()}
        region_coord_point = region_coord.cell(0).point
        region = region_map[original_regions[region_coord_point]]
        if checked_cube.shape[1] != 3 or checked_cube.shape[0] == 3:
            raise AssertionError("cube data shape isn't as expected")

        checked_cube.attributes["variable"] = checked_cube.var_name
        checked_cube.attributes["variable_standard_name"] = checked_cube.standard_name
        checked_cube.attributes["region"] = region
        if checked_cube.attributes["source_id"].startswith("UoM-CMIP"):
            scenario = "historical"
            model = "unspecified"
        else:
            scenario = "-".join(
                "ssp{}".format(
                    checked_cube.attributes["source_id"].split("ssp")[1]
                ).split("-")[:-3]
            )
            model = (
                checked_cube.attributes["source_id"]
                .split("-ssp")[0]
                .replace("UoM-", "")
            )
        checked_cube.attributes["scenario"] = scenario
        checked_cube.attributes["model"] = model
        checked_cube.attributes["climate_model"] = "MAGICC7"
        checked_cube.attributes["member_id"] = "unspecified"

        helper_region = ScmCube()
        helper_region.cube = checked_cube[:, region_coord_point]
        helper_region.cube.remove_coord(region_coord_name)
        timeseries_cubes[region] = helper_region

    output = (
        helper_region.convert_scm_timeseries_cubes_to_openscmdata(timeseries_cubes)
        .timeseries()
        .reset_index()
    )
    output["unit"] = output["unit"].map(unit_map)
    output["model"] = model
    output = ScmDataFrame(output)

    return output



"""
Module for wrangling netCDF-SCM netCDF files into other formats
"""
import os
import re

import click
import pandas as pd
from scmdata import ScmDataFrame, df_append

from .cli_parallel import _apply_func
from .cli_utils import (
    _find_dirs_meeting_func,
    _get_openscmdf_metadata_header,
    _get_outfile_dir,
    _write_ascii_file,
)
from .io import get_scmcube_helper, load_scmdataframe
from .wranglers import convert_scmdf_to_tuningstruc


def _tuningstrucs_blended_model_wrangling(  # pylint:disable=too-many-arguments
    logger, src, dst, regexp, force, drs, prefix
):
    regexp_compiled = re.compile(regexp)
    considered_regexps = []
    for dirpath, _, filenames in os.walk(src):
        if filenames:
            if not regexp_compiled.match(dirpath):
                continue

            if considered_regexps:
                if any([r.match(dirpath) for r in considered_regexps]):
                    continue

            regexp_here = _get_blended_model_regexp(drs, dirpath)
            logger.info("Wrangling %s", regexp_here)

            _tuningstrucs_blended_model_wrangling_inner_loop(
                src, regexp_here, dst, force, prefix
            )
            considered_regexps.append(regexp_here)


def _get_blended_model_regexp(drs, dirpath):
    scmcube = get_scmcube_helper(drs)
    ids = {
        k: v
        if any(
            [s in k for s in ["variable", "experiment", "activity", "mip", "member_id"]]
        )
        else ".*"
        for k, v in scmcube.process_path(dirpath).items()
    }
    for name, value in ids.items():
        setattr(scmcube, name, value)

    return re.compile("{}.*".format(scmcube.get_data_directory()))


def _tuningstrucs_blended_model_wrangling_inner_loop(
    src, regexp_inner, dst, force, prefix
):
    collected = []
    for dirpath_inner, _, filenames_inner in os.walk(src):
        if filenames_inner:
            if not regexp_inner.match(dirpath_inner):
                continue

            openscmdf = df_append(
                [
                    load_scmdataframe(os.path.join(dirpath_inner, f))
                    for f in filenames_inner
                ]
            )
            tmp_ts = openscmdf.timeseries().reset_index()
            tmp_ts["unit"] = tmp_ts["unit"].astype(str)
            openscmdf = ScmDataFrame(tmp_ts)

            collected.append(openscmdf)

    convert_scmdf_to_tuningstruc(df_append(collected), dst, force=force, prefix=prefix)


def _do_wrangling(  # pylint:disable=too-many-arguments
    logger,
    src,
    dst,
    regexp,
    out_format,
    force,
    wrangle_contact,
    drs,
    number_workers,
    target_units_specs,
):
    regexp_compiled = re.compile(regexp)
    if target_units_specs is not None:
        target_units_specs = pd.read_csv(target_units_specs)

    if out_format in (
        "mag-files",
        "mag-files-average-year-start-year",
        "mag-files-average-year-mid-year",
        "mag-files-average-year-end-year",
        "mag-files-point-start-year",
        "mag-files-point-mid-year",
        "mag-files-point-end-year",
        "magicc-input-files",
        "magicc-input-files-average-year-start-year",
        "magicc-input-files-average-year-mid-year",
        "magicc-input-files-average-year-end-year",
        "magicc-input-files-point-start-year",
        "magicc-input-files-point-mid-year",
        "magicc-input-files-point-end-year",
    ):
        _do_magicc_wrangling(
            logger,
            src,
            dst,
            regexp_compiled,
            out_format,
            force,
            wrangle_contact,
            drs,
            number_workers,
            target_units_specs,
        )
    else:  # pragma: no cover # emergency valve (should be caught by click on call)
        raise ValueError("Unsupported format: {}".format(out_format))


def _do_magicc_wrangling(  # pylint:disable=too-many-arguments,too-many-locals
    logger,
    src,
    dst,
    regexp_compiled,
    out_format,
    force,
    wrangle_contact,
    drs,
    number_workers,
    target_units_specs,
):
    crunch_list, failures_dir_finding = _find_dirs_meeting_func(
        src, regexp_compiled.match
    )

    if out_format in (
        "mag-files",
        "mag-files-average-year-start-year",
        "mag-files-average-year-mid-year",
        "mag-files-average-year-end-year",
        "mag-files-point-start-year",
        "mag-files-point-mid-year",
        "mag-files-point-end-year",
        "magicc-input-files",
        "magicc-input-files-average-year-start-year",
        "magicc-input-files-average-year-mid-year",
        "magicc-input-files-average-year-end-year",
        "magicc-input-files-point-start-year",
        "magicc-input-files-point-mid-year",
        "magicc-input-files-point-end-year",
    ):
        failures_wrangling = _apply_func(
            logger,
            _wrangle_magicc_files,
            [{"fnames": f, "dpath": d} for d, f in crunch_list],
            common_kwarglist={
                "logger": logger,
                "dst": dst,
                "force": force,
                "out_format": out_format,
                "target_units_specs": target_units_specs,
                "wrangle_contact": wrangle_contact,
                "drs": drs,
            },
            n_workers=number_workers,
            style="processes",
        )

    else:  # pragma: no cover # emergency valve
        raise AssertionError(
            "how did we get here, click should have prevented the --out-format "
            "option..."
        )

    if failures_dir_finding or failures_wrangling:
        raise click.ClickException(
            "Some files failed to process. See the logs for more details"
        )


def _wrangle_magicc_files(  # pylint:disable=too-many-arguments
    logger,
    fnames,
    dpath,
    dst,
    force,
    out_format,
    target_units_specs,
    wrangle_contact,
    drs,
):
    logger.info("Attempting to process: %s", fnames)
    openscmdf, metadata, header = _get_openscmdf_metadata_header(
        fnames, dpath, target_units_specs, wrangle_contact, out_format
    )

    outfile_dir = _get_outfile_dir(dpath, drs, dst)

    _write_ascii_file(
        openscmdf, metadata, header, outfile_dir, fnames, force, out_format, drs,
    )



import warnings

from netcdf_scm.errors import raise_no_iris_warning


def test_raise_no_iris_warning():
    expected_warn = (
        "A compatible version of Iris is not installed, not all functionality will "
        "work. We recommend installing the lastest version of Iris using conda to "
        "address this."
    )

    with warnings.catch_warnings(record=True) as no_iris_warning:
        raise_no_iris_warning()

    assert len(no_iris_warning) == 1
    assert str(no_iris_warning[0].message) == expected_warn



"""
Module for the normaliser which only removes drift in the pre-industrial control run (drift is calculated using a running-mean)
"""

import logging

from .running_mean import NormaliserRunningMean

LOGGER = logging.getLogger(__name__)


class NormaliserRunningMeanDedrift(NormaliserRunningMean):
    """
    Normaliser which calculates drift in the pre-industrial control using a running mean

    Each normalisation value is the change in a 21-year mean with respect to the running mean at the branch point.
    Each point is centred on the equivalent point in the pre-industrial control simulation. This means that the
    reference values are always zero in their first timestep.

    If there is insufficient data to create a full 21-year window at the edge of the simulation then a linear
    extrapolation of the running-mean is used to extend the normalisation values to cover the required
    full range.

    At present, only a 21-year running mean is implemented.
    """

    def __init__(self):
        self._method_name = "21-yr-running-mean-dedrift"

    @staticmethod
    def _get_reference_values(indata, picontrol, picontrol_branching_time):
        res = NormaliserRunningMean._get_reference_values(
            indata, picontrol, picontrol_branching_time
        )

        time_point_first = res.columns.min()
        res = res.subtract(res.loc[:, time_point_first], axis=0)

        return res



import numpy.testing as npt
from openscm_units import unit_registry

from netcdf_scm.io import load_mag_file


def test_stitch_output(stitched_mag_file):
    drs = "CMIP6Output" if "CMIP6" in stitched_mag_file else "MarbleCMIP5"
    res = load_mag_file(stitched_mag_file, drs=drs)

    check_unit = 3 * unit_registry(res.get_unique_meta("unit", no_duplicates=True))

    assert str(check_unit.units) != ""
    npt.assert_equal(check_unit.magnitude, 3)



import datetime as dt
import re
from unittest.mock import patch

import pytest
import requests
from scmdata.run import ScmRun

from netcdf_scm.citing import check_license_cmip5, check_license_cmip6
from netcdf_scm.citing.cmip5 import _get_cmip5_citation_key, get_cmip5_bibtex_citation
from netcdf_scm.citing.cmip6 import (
    _get_cmip6_citation_key,
    get_cmip6_bibtex_citation,
    get_license_info,
)
from netcdf_scm.errors import NoLicenseInformationError, NonStandardLicenseError


@pytest.fixture
def test_scmdata():
    out = ScmRun(
        data=[1, 2, 3],
        index=[10, 20, 30],
        columns={
            "model": "junk",
            "scenario": "junk",
            "variable": "junk",
            "region": "junk",
            "unit": "junk",
        },
    )

    out.metadata = {"junk key": "junk value"}

    return out


def test_get_license_info_single_key(test_scmdata):
    tlicense_info = "license information here"
    test_scmdata.metadata["license"] = tlicense_info

    res = get_license_info(test_scmdata)

    assert res == [tlicense_info]


def test_get_license_info_two_keys(test_scmdata):
    tlicense_info_1 = "license information here"
    tlicense_info_2 = "license information here"
    test_scmdata.metadata["(child) license"] = tlicense_info_1
    test_scmdata.metadata["(normalisation) license"] = tlicense_info_2

    res = get_license_info(test_scmdata)

    assert set(res) == {tlicense_info_1, tlicense_info_2}


def test_get_license_info_missing_license(test_scmdata):
    error_msg = re.escape("Metadata found: {}".format(test_scmdata.metadata))
    with pytest.raises(NoLicenseInformationError, match=error_msg):
        get_license_info(test_scmdata)


def test_check_license_missing_license(test_scmdata):
    error_msg = re.escape("Metadata found: {}".format(test_scmdata.metadata))
    with pytest.raises(NoLicenseInformationError, match=error_msg):
        check_license_cmip6(test_scmdata)


VALID_LICENSE_TXT = (
    "Creative Commons Attribution ShareAlike 4.0 International License. "
    "See website for terms of use governing CMIP6 output, including "
    "citation requirements and proper acknowledgment"
)


@pytest.mark.parametrize(
    "txt,exp_error",
    (
        (VALID_LICENSE_TXT, False),
        (
            "Creative Commons Attribution[]-ShareAlike 4.0 International License. "
            "See website for terms of use governing CMIP6 output, including "
            "citation requirements and proper acknowledgment",
            False,
        ),
        (
            "Creative Commons Attribution[]-ShareAlike 4.0 Non-Commercial License. "
            "See website for terms of use governing CMIP6 output, including "
            "citation requirements and proper acknowledgment",
            True,
        ),
        ("Nonsense", True),
    ),
)
def test_check_license_single_key(test_scmdata, txt, exp_error):
    test_scmdata.metadata["license"] = txt
    if exp_error:
        error_msg = re.escape("Non-standard licenses: {}".format({txt}))
        with pytest.raises(NonStandardLicenseError, match=error_msg):
            check_license_cmip6(test_scmdata)

    else:
        check_license_cmip6(test_scmdata)


@pytest.mark.parametrize(
    "txt_1,txt_2,exp_error_1,exp_error_2",
    (
        (VALID_LICENSE_TXT, VALID_LICENSE_TXT, False, False),
        ("Nonsense", VALID_LICENSE_TXT, True, False),
        (
            VALID_LICENSE_TXT,
            VALID_LICENSE_TXT.replace("International", "NonCommercial International"),
            False,
            True,
        ),
        (
            VALID_LICENSE_TXT,
            VALID_LICENSE_TXT.replace("Attribution", "Attribution-NonCommercial"),
            False,
            True,
        ),
        ("Nonsense", "More nonsense", True, True),
    ),
)
def test_check_license_multiple_keys(
    test_scmdata, txt_1, txt_2, exp_error_1, exp_error_2
):
    test_scmdata.metadata["(child) license"] = txt_1
    test_scmdata.metadata["(normalisation) license"] = txt_2
    if exp_error_1 or exp_error_2:
        if exp_error_1 and exp_error_2:
            error_msg = re.escape("Non-standard licenses: {}".format({txt_1, txt_2}))
        elif exp_error_1:
            error_msg = re.escape("Non-standard licenses: {}".format({txt_1}))
        else:
            error_msg = re.escape("Non-standard licenses: {}".format({txt_2}))

        with pytest.raises(NonStandardLicenseError, match=error_msg):
            check_license_cmip6(test_scmdata)

    else:
        check_license_cmip6(test_scmdata)


@pytest.mark.parametrize(
    "model_name,restricted",
    (
        ("ACCESS1.0", False),
        ("INM-CM4", False),
        ("MIROC-ESM", True),
        ("MIROC-ESM-CHEM", True),
        ("MRI-CGCM3", True),
        ("NICAM.09", True),
    ),
)
def test_check_license_cmip5(test_scmdata, model_name, restricted):
    test_scmdata["climate_model"] = model_name

    if restricted:
        error_msg = re.escape(
            "Output is only available for non-commercial research and "
            "education purposes, see "
            "`<https://pcmdi.llnl.gov/mips/cmip5/docs/CMIP5_modeling_groups.pdf>`_"
        )

        with pytest.raises(NonStandardLicenseError, match=error_msg):
            check_license_cmip5(test_scmdata)

    else:
        check_license_cmip5(test_scmdata)


@patch("netcdf_scm.citing.cmip5._query_cmip5_doi")
def test_get_cmip5_bibtex_citation(mock_query_cmip5_doi):
    raw_response = "@misc{https://doi.org/10.1594/wdcc/cmip5.csa0c1,\n  doi = {10.1594/WDCC/CMIP5.CSA0C1},\n  url = {http://cera-www.dkrz.de/WDCC/CMIP5/Compact.jsp?acronym=CSA0c1},\n  author = {Bi, Dave and Dix, Martin and Marsland, Simon and O'Farrell, Siobhan and Uotila, Petteri and Hirst, Tony and Kowalczyk, Eva and Rashid, Harun and Sun, Zhian and Collier, Mark and Dommenget, Katja and Golebiewski, Maciej and Hannah, Nicholas and Fiedler, Russell and Franklin, Charmaine and Lewis, Sophie and Ma, Yimin and Petrelli, Paola and Stevens, Lauren and Sullivan, Arnold and Uhe, Peter and Vohralik, Peter and Watterson, Ian and Yan, Hailin and Zhou, Xiaobing},\n  keywords = {Climate, ACCESS1-0, CMIP5, IPCC, IPCC-AR5, IPCC-AR5_CMIP5, IPCC-DDC, climate simulation},\n  language = {en},\n  title = {ACCESS1-0 model output prepared for CMIP5 1pctCO2, served by ESGF},\n  publisher = {World Data Center for Climate (WDCC) at DKRZ},\n  year = {2016}\n}\n"

    mock_query_cmip5_doi.return_value = raw_response

    res = get_cmip5_bibtex_citation("ACCESS1.0", "1pctCO2")

    assert res == raw_response.replace(
        "https://doi.org/10.1594/wdcc/cmip5.csa0c1", "cmip5data_ACCESS1.0_1pctCO2"
    )


def test_get_cmip5_bibtex_citation_failure():
    model = "ACCESS1.0"
    scenario = "junk"
    error_msg = re.escape("No CMIP5 data for {} {}".format(model, scenario))
    with pytest.raises(ValueError, match=error_msg):
        get_cmip5_bibtex_citation(model, scenario)


@pytest.mark.parametrize("version", (None, "YYYYMMDD"))
@patch("netcdf_scm.citing.cmip6._get_cmip6_doi")
@patch("netcdf_scm.citing.cmip6._query_cmip6")
def test_get_cmip6_bibtex_citation(mock_query_cmip6, mock_get_cmip6_doi, version):
    mip_era = "CMIP6"
    activity_id = "ScenarioMIP"
    institute_id = "CSIRO-ARCCSS"
    climate_model = "ACCESS-CM2"
    scenario = "ssp585"

    drs_id = ".".join([mip_era, activity_id, institute_id, climate_model, scenario])

    mock_doi = "10.22033/ESGF/CMIP6.4332"
    mock_get_cmip6_doi.return_value = mock_doi

    raw_response = "@misc{https://doi.org/10.22033/esgf/cmip6.4332,\n  doi = {10.22033/ESGF/CMIP6.4332},\n  url = {http://cera-www.dkrz.de/WDCC/meta/CMIP6/CMIP6.ScenarioMIP.CSIRO-ARCCSS.ACCESS-CM2.ssp585},\n  author = {Dix, Martin and Bi, Doahua and Dobrohotoff, Peter and Fiedler, Russell and Harman, Ian and Law, Rachel and Mackallah, Chloe and Marsland, Simon and O'Farrell, Siobhan and Rashid, Harun and Srbinovsky, Jhan and Sullivan, Arnold and Trenham, Claire and Vohralik, Peter and Watterson, Ian and Williams, Gareth and Woodhouse, Matthew and Bodman, Roger and Dias, Fabio Boeira and Domingues, Catia and Hannah, Nicholas and Heerdegen, Aidan and Savita, Abhishek and Wales, Scott and Allen, Chris and Druken, Kelsey and Evans, Ben and Richards, Clare and Ridzwan, Syazwan Mohamed and Roberts, Dale and Smillie, Jon and Snow, Kate and Ward, Marshall and Yang, Rui},\n  keywords = {CMIP6, climate, CMIP6.ScenarioMIP.CSIRO-ARCCSS.ACCESS-CM2.ssp585},\n  language = {en},\n  title = {CSIRO-ARCCSS ACCESS-CM2 model output prepared for CMIP6 ScenarioMIP ssp585},\n  publisher = {Earth System Grid Federation},\n  year = {2019}\n}\n"

    mock_query_cmip6.return_value = raw_response

    res = get_cmip6_bibtex_citation(
        mip_era, activity_id, institute_id, climate_model, scenario, version=version
    )

    mock_get_cmip6_doi.assert_called_with(drs_id)
    query_val = "https://doi.org/{}".format(mock_doi)
    mock_query_cmip6.assert_called_with(query_val)

    if version is None:
        expected_version = dt.date.today().strftime("%Y%m%d")
    else:
        expected_version = version

    assert res == raw_response.replace(
        query_val.lower(), "cmip6data_{}".format(drs_id.replace(".", "_"))
    ).replace(
        "CSIRO-ARCCSS ACCESS-CM2 model output prepared for CMIP6 ScenarioMIP ssp585",
        "CSIRO-ARCCSS ACCESS-CM2 model output prepared for CMIP6 ScenarioMIP ssp585. Version {}".format(
            expected_version
        ),
    )


@patch("netcdf_scm.citing.cmip6._get_cmip6_doi")
def test_get_cmip6_bibtex_citation_failure(mock_get_cmip6_doi):
    mock_get_cmip6_doi.side_effect = requests.exceptions.HTTPError("test message")
    mip_era = "CMIP6"
    activity_id = "ScenarioMIP"
    institute_id = "CSIRO-ARCCSS"
    climate_model = "ACCESS-CM2"
    scenario = "junk"

    drs_id = ".".join([mip_era, activity_id, institute_id, climate_model, scenario])

    error_msg = re.escape(
        "No citation could be found for {}.\nError was: test message".format(drs_id)
    )
    with pytest.raises(ValueError, match=error_msg):
        get_cmip6_bibtex_citation(
            mip_era, activity_id, institute_id, climate_model, scenario
        )


def test_get_citation_key_cmip5():
    model = "model"
    scenario = "scenario"
    exp = "cmip5data_{}_{}".format(model, scenario)
    res = _get_cmip5_citation_key(model, scenario)

    assert res == exp


def test_get_citation_key_cmip6():
    mip_era = "mip_era"
    activity_id = "activity_id"
    institute_id = "institute_id"
    model = "model"
    scenario = "scenario"
    exp = "cmip6data_{}_{}_{}_{}_{}".format(
        mip_era, activity_id, institute_id, model, scenario
    )
    res = _get_cmip6_citation_key(mip_era, activity_id, institute_id, model, scenario)

    assert res == exp



import re

import numpy as np
import pytest
from scmdata import ScmDataFrame

from netcdf_scm.stitching import (
    get_branch_time,
    get_parent_replacements,
    step_up_family_tree,
)


@pytest.fixture(scope="function")
def junk_data():
    out = ScmDataFrame(
        np.arange(100),
        index=np.arange(1850, 1950),
        columns={
            "scenario": ["rcp26"],
            "experiment_id": ["rcp26"],
            "model": ["model1"],
            "climate_model": ["climate_model1"],
            "mip_era": ["CMIP6"],
            "region": ["World"],
            "variable": ["tas"],
            "unit": ["K"],
        },
    )

    out.metadata = {
        "branch_time_in_parent": 0.0,
        "parent_time_units": "days since 1850-01-01",
        "calendar": "365_day",
        "source_id": "UoM mocked data",
        "experiment_id": "rcp26",
        "netcdf-scm crunched file": "/path/to/mocked/rcp26/data_rcp26.nc",
    }

    return out


@pytest.fixture
def junk_data_picontrol(junk_data):
    out = junk_data.copy()
    out.set_meta("piControl", "scenario")
    out.set_meta("piControl", "experiment_id")
    out.metadata = {
        "other": "metadata",
        "goes": "here",
        "branch_time_in_parent": -1001.1,
        "parent_time_units": "days since 1850-01-01",
        "calendar": "365_day",
        "source_id": "UoM mocked data",
        "experiment_id": "piControl",
        "netcdf-scm crunched file": "/path/to/mocked/piControl/data_piControl.nc",
    }

    return out


def test_get_branch_time_cmip5_not_parent_error(junk_data):
    junk_data.set_meta("CMIP5", "mip_era")

    error_msg = re.escape(
        "CMIP5 data does not contain information about the branch time "
        "in the child's time axis"
    )
    with pytest.raises(ValueError, match=error_msg):
        get_branch_time(junk_data, parent=False)


def test_get_branch_time_cmip5_no_parent_path_error(junk_data):
    junk_data.set_meta("CMIP5", "mip_era")

    error_msg = re.escape("If using cmip5 data, you must provide `parent_path`")
    with pytest.raises(ValueError, match=error_msg):
        get_branch_time(junk_data)


def test_get_branch_time_cmip6_child_no_source_path_error(junk_data):
    error_msg = re.escape(
        "If not requesting parent data, you must provide `source_path`"
    )
    with pytest.raises(ValueError, match=error_msg):
        get_branch_time(junk_data, parent=False)


@pytest.mark.parametrize(
    "start,expected",
    (
        ("(child)", "(parent)"),
        ("(parent)", "(grandparent)"),
        ("(grandparent)", "(greatgrandparent)"),
        ("(greatgrandparent)", "(greatgreatgrandparent)"),
        ("(greatgreatgreatgreatgrandparent)", "(greatgreatgreatgreatgreatgrandparent)"),
    ),
)
def test_step_up_family_tree(start, expected):
    assert step_up_family_tree(start) == expected


@pytest.mark.parametrize(
    "start,expected,cmip5",
    (
        (
            {
                "parent_activity_id": "CMIP",
                "parent_experiment_id": "historical",
                "parent_mip_era": "CMIP6",
                "parent_source_id": "UoM",
                "parent_variant_label": "r1i1p1f1",
                "other_meta": "something",
            },
            {
                "parent_activity_id": "CMIP",
                "parent_experiment_id": "historical",
                "parent_mip_era": "CMIP6",
                "parent_source_id": "UoM",
                "parent_member_id": "r1i1p1f1",
            },
            False,
        ),
        (
            {
                "parent_activity_id": "C M I P",
                "parent_experiment_id": "p i C o n t r o l",
                "parent_mip_era": "C M I P 6",
                "parent_source_id": "C N R M - C M6-1-H R",
                "parent_variant_label": "r 1i1p1f 2",
            },
            {
                "parent_activity_id": "CMIP",
                "parent_experiment_id": "piControl",
                "parent_mip_era": "CMIP6",
                "parent_source_id": "CNRM-CM6-1-HR",
                "parent_member_id": "r1i1p1f2",
            },
            False,
        ),
        (
            {
                "parent_experiment": "piControl",
                "parent_experiment_id": "piControl",
                "parent_experiment_rip": "r1i1p1",
                "other_meta": "something",
            },
            {
                "parent_experiment": "piControl",
                "parent_experiment_id": "piControl",
                "parent_ensemble_member": "r1i1p1",
            },
            True,
        ),
    ),
)
def test_get_parent_replacements(junk_data, start, expected, cmip5):
    junk_data.metadata = start

    if cmip5:
        junk_data.set_meta("CMIP5", "mip_era")
    else:
        junk_data.set_meta("CMIP6", "mip_era")

    assert get_parent_replacements(junk_data) == expected


def test_get_parent_replacements_no_rip_error(junk_data):
    junk_data.set_meta("CMIP5", "mip_era")
    error_msg = re.escape("No `parent_experiment_rip` in metadata")
    with pytest.raises(KeyError, match=error_msg):
        get_parent_replacements(junk_data)


def test_get_parent_replacements_no_parent_variant_label_error(junk_data):
    junk_data.set_meta("CMIP6", "mip_era")
    error_msg = re.escape("No `parent_variant_label` in metadata")
    with pytest.raises(KeyError, match=error_msg):
        get_parent_replacements(junk_data)



from base_normalisation_unit import NormaliserTester

from netcdf_scm.normalisation import AfterBranchTimeMean


class TestRunningMean(NormaliserTester):
    tclass = AfterBranchTimeMean

    def test_method_name(self):
        assert self.tclass().method_name == "31-yr-mean-after-branch-time"



import datetime
import re
import warnings
from unittest.mock import patch

import cf_units
import dask.array as da
import iris
import numpy as np
import numpy.ma as ma
import numpy.testing as npt
import pytest

from netcdf_scm.utils import (
    _assert_only_cube_dim_coord_is_time,
    apply_mask,
    assert_all_time_axes_same,
    broadcast_onto_lat_lon_grid,
    get_cube_timeseries_data,
    get_scm_cube_time_axis_in_calendar,
    take_lat_lon_mean,
    unify_lat_lon,
)


def test_assert_only_cube_dim_coord_is_time(test_generic_tas_cube):
    original_cube = test_generic_tas_cube.cube

    error_msg = re.escape("Should only have time coordinate here")

    with pytest.raises(AssertionError, match=error_msg):
        _assert_only_cube_dim_coord_is_time(test_generic_tas_cube)

    # can safely ignore these warnings here
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", ".*Collapsing spatial coordinate 'latitude' without weighting.*"
        )
        test_generic_tas_cube.cube = original_cube.collapsed(
            ["latitude", "time"], iris.analysis.MEAN
        )

    with pytest.raises(AssertionError, match=error_msg):
        _assert_only_cube_dim_coord_is_time(test_generic_tas_cube)

    # can safely ignore these warnings here
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", ".*Using DEFAULT_SPHERICAL.*")
        test_generic_tas_cube.cube = original_cube.collapsed(
            ["latitude", "longitude"],
            iris.analysis.MEAN,
            weights=iris.analysis.cartography.area_weights(original_cube),
        )

    _assert_only_cube_dim_coord_is_time(test_generic_tas_cube)


@patch("netcdf_scm.utils._assert_only_cube_dim_coord_is_time")
def test_get_cube_timeseries_data(mock_assert_only_time, test_generic_tas_cube):
    expected = test_generic_tas_cube.cube.data
    result = get_cube_timeseries_data(test_generic_tas_cube)

    np.testing.assert_array_equal(result, expected)
    mock_assert_only_time.assert_called_with(test_generic_tas_cube)


@pytest.mark.parametrize("out_calendar", ["gregorian", "julian", "365_day"])
def test_get_cube_time_axis_in_calendar(test_generic_tas_cube, out_calendar):
    tcn = test_generic_tas_cube.cube.coord_dims("time")[0]
    ttime = test_generic_tas_cube.cube.dim_coords[tcn]
    expected = cf_units.num2date(ttime.points, ttime.units.name, out_calendar)

    result = get_scm_cube_time_axis_in_calendar(test_generic_tas_cube, out_calendar)
    np.testing.assert_array_equal(result, expected)


def test_assert_all_time_axes_same(test_generic_tas_cube):
    tcn = test_generic_tas_cube.cube.coord_dims("time")[0]
    ttime = test_generic_tas_cube.cube.dim_coords[tcn]
    ttime_axis = cf_units.num2date(ttime.points, ttime.units.name, "gregorian")

    assert_all_time_axes_same([ttime_axis, ttime_axis])

    otime_axis = ttime_axis - datetime.timedelta(10)

    error_msg = re.escape("all the time axes should be the same")
    with pytest.raises(AssertionError, match=error_msg):
        assert_all_time_axes_same([otime_axis, ttime_axis])


def test_take_lat_lon_mean(test_generic_tas_cube):
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", ".*Using DEFAULT_SPHERICAL.*")
        tweights = iris.analysis.cartography.area_weights(test_generic_tas_cube.cube)

    assert len(test_generic_tas_cube.cube.dim_coords) == 3

    result, normalising_factor = take_lat_lon_mean(test_generic_tas_cube, tweights)

    assert len(result.cube.dim_coords) == 1
    assert result.cube.cell_methods[0].method == "mean"
    assert result.cube.cell_methods[0].coord_names == ("time",)
    assert result.cube.cell_methods[1].method == "mean"
    assert result.cube.cell_methods[1].coord_names == ("latitude", "longitude")

    npt.assert_allclose(normalising_factor, 5.09499402e14)


@pytest.mark.parametrize(
    "weight_func,expected",
    (
        (lambda x: x >= 10, np.sum(range(10))),
        (lambda x: x >= 6, np.sum(range(6))),
        (lambda x: x < 6, np.sum(range(6, 12))),
    ),
)
@pytest.mark.parametrize("dask_data", (True, False))
def test_take_lat_lon_mean_with_mask(test_all_cubes, weight_func, expected, dask_data):
    lat_lon_slice = test_all_cubes._get_lat_lon_slice()
    weights = np.arange(np.prod(lat_lon_slice.shape)).reshape(lat_lon_slice.shape)
    mask = broadcast_onto_lat_lon_grid(
        test_all_cubes, weight_func(weights).astype(bool)
    )

    if dask_data:
        test_all_cubes.cube.data = da.ma.masked_array(
            data=test_all_cubes.cube.data, mask=mask
        )
    else:
        test_all_cubes.cube.data = ma.masked_array(
            data=test_all_cubes.cube.data, mask=mask
        )

    result, normalising_factor = take_lat_lon_mean(test_all_cubes, weights)

    assert len(result.cube.dim_coords) == 1
    assert result.cube.cell_methods[0].method == "mean"
    assert result.cube.cell_methods[0].coord_names == ("latitude", "longitude")

    npt.assert_allclose(normalising_factor, expected)


@pytest.mark.parametrize("lazy", [True, False])
def test_apply_mask(test_generic_tas_cube, lazy):
    tmask = np.full(test_generic_tas_cube.cube.shape, True)

    assert test_generic_tas_cube.cube.has_lazy_data()
    if not lazy:
        np.testing.assert_equal(test_generic_tas_cube.cube.data.mask, ~tmask)

    result = apply_mask(test_generic_tas_cube, tmask)
    if lazy:
        assert result.cube.has_lazy_data()
    else:
        assert not result.cube.has_lazy_data()
    np.testing.assert_equal(result.cube.data.mask, tmask)


@pytest.mark.parametrize(
    "ttol",
    [0.1, 10 ** -3, 1.5 * 10 ** -5, 10 ** -5, 0.9 * 10 ** -5, 10 ** -10, "default"],
)
def test_unify_lat_lon(test_generic_tas_cube, ttol):
    def get_starting_list(scale):
        base_cube = test_generic_tas_cube.cube.copy()
        other_cube = base_cube.copy()
        other_cube.coords("longitude")[0].points = (
            scale * other_cube.coords("longitude")[0].points
        )
        other_cube.coords("latitude")[0].points = (
            scale * other_cube.coords("latitude")[0].points
        )

        assert (
            other_cube.coords("longitude")[0].points
            == scale * base_cube.coords("longitude")[0].points
        ).all()
        assert (
            other_cube.coords("latitude")[0].points
            == scale * base_cube.coords("latitude")[0].points
        ).all()

        return iris.cube.CubeList([base_cube, other_cube])

    default = False if ttol != "default" else True
    ttol = ttol if ttol != "default" else 10 ** -6
    for loop_ttol in [ttol * 0.1, ttol * 10 ** -5]:
        tlist = get_starting_list(scale=1 + loop_ttol)
        if default:
            unify_lat_lon(tlist)
        else:
            unify_lat_lon(tlist, rtol=ttol)

        assert (
            tlist[0].coords("longitude")[0].points
            == tlist[1].coords("longitude")[0].points
        ).all()
        assert (
            tlist[0].coords("latitude")[0].points
            == tlist[1].coords("latitude")[0].points
        ).all()

    error_msg = re.escape(
        "Cannot unify latitude and longitude, relative difference in co-ordinates "
        "is greater than {}".format(ttol)
    )
    with pytest.raises(ValueError, match=error_msg):
        tlist = get_starting_list(ttol * 1.01)
        if default:
            unify_lat_lon(tlist)
        else:
            unify_lat_lon(tlist, rtol=ttol)


@pytest.mark.parametrize("lazy", [True, False])
@pytest.mark.parametrize("transpose", [True, False])
@patch.object(iris.cube.Cube, "has_lazy_data")
def test_broadcast_onto_lat_lon_grid(
    mock_has_lazy_data, test_generic_tas_cube, test_sftlf_file, lazy, transpose
):
    original = iris.load_cube(test_sftlf_file).data
    if transpose:
        ar_to_broadcast = np.transpose(original)
    else:
        ar_to_broadcast = original

    mock_has_lazy_data.return_value = lazy

    res = broadcast_onto_lat_lon_grid(test_generic_tas_cube, ar_to_broadcast)

    assert isinstance(res, da.Array) if lazy else isinstance(res, np.ndarray)
    expected = np.broadcast_to(original, test_generic_tas_cube.cube.shape)

    np.testing.assert_allclose(res, expected)


def test_broadcast_onto_lat_lon_grid_errors(test_generic_tas_cube, test_sftlf_file):
    ar_to_broadcast = iris.load_cube(test_sftlf_file).data[1:, 1:]
    error_msg = re.escape(
        "the ``array_in`` must be the same shape as the "
        "cube's longitude-latitude grid"
    )
    with pytest.raises(AssertionError, match=error_msg):
        broadcast_onto_lat_lon_grid(test_generic_tas_cube, ar_to_broadcast)



"""netCDF-SCM's custom error handling"""
import warnings


def raise_no_iris_warning():
    """
    Raise a warning that iris is not installed

    Warns
    -----
    UserWarning
        `Iris <https://github.com/SciTools/iris>`_ is not installed
    """
    warn_msg = (
        "A compatible version of Iris is not installed, not all functionality will "
        "work. We recommend installing the lastest version of Iris using conda to "
        "address this."
    )
    warnings.warn(warn_msg)


class NoLicenseInformationError(AttributeError):
    """
    Exception raised when a dataset contains no license information
    """


class NonStandardLicenseError(ValueError):
    """
    Exception raised when a dataset contains a non-standard license

    For example, if a CMIP6 dataset does not contain a Creative Commons
    Attribution ShareAlike 4.0 International License
    """



"""
netCDF-SCM, a package for preparing netCDF files for use in Simple Climate Models (SCMs)

The package builds on the `iris <https://github.com/SciTools/iris>`_ package.
"""
import logging

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

logging.getLogger(__name__).addHandler(logging.NullHandler())



import datetime as dt
import json
from glob import glob
from os import walk
from os.path import isdir, isfile, join
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest
from click.testing import CliRunner

import netcdf_scm
from netcdf_scm.cli import crunch_data
from netcdf_scm.io import load_scmdataframe
from netcdf_scm.iris_cube_wrappers import ScmCube


def test_crunching_against_knmi(
    tmpdir, caplog, test_data_knmi_dir, test_data_marble_cmip5_dir
):
    INPUT_DIR = test_data_marble_cmip5_dir
    OUTPUT_DIR = str(tmpdir)
    VAR_TO_CRUNCH = ".*tas.*"
    crunch_contact = "knmi-verification"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            crunch_data,
            [
                INPUT_DIR,
                OUTPUT_DIR,
                crunch_contact,
                "--drs",
                "MarbleCMIP5",
                "--regexp",
                VAR_TO_CRUNCH,
                "-f",
                "--small-number-workers",
                1,
            ],
        )
    assert result.exit_code == 0
    assert "netcdf-scm: {}".format(netcdf_scm.__version__) in caplog.messages
    assert (
        "Making output directory: {}/netcdf-scm-crunched".format(OUTPUT_DIR)
        in caplog.messages
    )

    # Check that there is a log file  which contains 'INFO' log messages
    log_fnames = glob(join(OUTPUT_DIR, "netcdf-scm-crunched", "*.log"))
    assert len(log_fnames) == 1

    with open(log_fnames[0]) as fh:
        log_file = fh.read()
        assert "DEBUG" in log_file
    # Check that the logs are also written to stderr
    assert "DEBUG" not in result.stderr
    assert "INFO" in result.stderr

    # Check the output_tracker file
    with open(
        join(OUTPUT_DIR, "netcdf-scm-crunched", "netcdf-scm_crunched.jsonl")
    ) as fh:
        lines = fh.readlines()
        assert len(lines) == 9

        # check that CanESM2 has areacella file
        for line in lines:
            d = json.loads(line)
            if "tas_Amon_CanESM2_1pctCO2_r1i1p1_189201-190312.nc" in d["files"][0]:
                checked_metadata = True
                assert len(d["metadata"]["areacella"]["files"]) == 1
                assert len(d["metadata"]["sftlf"]["files"]) == 1

    assert checked_metadata

    THRESHOLD_PERCENTAGE_DIFF = 10 ** -1
    files_found = 0
    for dirpath, dirnames, filenames in walk(OUTPUT_DIR):
        if not dirnames:
            assert len(filenames) == 1
            filename = filenames[0]
            files_found += 1

            knmi_data_name = "global_{}.dat".format("_".join(filename.split("_")[1:6]))
            knmi_data_path = join(test_data_knmi_dir, knmi_data_name)

            if not isfile(knmi_data_path):
                print("No data available for {}".format(knmi_data_path))
                continue

            knmi_data = pd.read_csv(
                knmi_data_path,
                skiprows=3,
                delim_whitespace=True,
                header=None,
                names=["year", 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
            ).melt(id_vars="year", var_name="month")
            knmi_data["year"] = knmi_data["year"].astype(int)
            knmi_data["month"] = knmi_data["month"].astype(int)
            knmi_data = knmi_data.set_index(["year", "month"])

            crunched_data = load_scmdataframe(join(dirpath, filename))
            assert crunched_data.metadata["crunch_contact"] == crunch_contact

            comparison_data = (
                crunched_data.filter(region="World")
                .timeseries()
                .stack()
                .to_frame()
                .reset_index()[["time", 0]]
            )
            comparison_data = comparison_data.rename({0: "value"}, axis="columns")

            comparison_data["year"] = comparison_data["time"].apply(lambda x: x.year)
            comparison_data["month"] = comparison_data["time"].apply(lambda x: x.month)

            comparison_data = comparison_data.drop("time", axis="columns")
            comparison_data = comparison_data.set_index(["year", "month"])

            rel_difference = (knmi_data - comparison_data) / knmi_data
            # drop regions where times are not equal
            rel_difference = rel_difference.dropna()
            assert not rel_difference.empty, "not testing anything"

            assert_message = "{} data is not the same to within {}%".format(
                filename, THRESHOLD_PERCENTAGE_DIFF
            )
            all_close = (
                np.abs(rel_difference.values) < THRESHOLD_PERCENTAGE_DIFF / 100
            ).all()
            assert all_close, assert_message

            print(
                "{} file matches KNMI data to within {}%".format(
                    filename, THRESHOLD_PERCENTAGE_DIFF
                )
            )

    assert files_found == 9


def test_crunching_join_files(tmpdir, caplog, test_data_cmip6output_dir):
    INPUT_DIR = join(
        test_data_cmip6output_dir,
        "CMIP6",
        "CMIP",
        "IPSL",
        "IPSL-CM6A-LR",
        "piControl",
        "r1i1p1f1",
        "Amon",
        "tas",
        "gr",
        "v20181123",
    )
    OUTPUT_DIR = str(tmpdir)
    crunch_contact = "join-files-test"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            crunch_data,
            [
                INPUT_DIR,
                OUTPUT_DIR,
                crunch_contact,
                "--drs",
                "CMIP6Output",
                "-f",
                "--small-threshold",
                0,
                "--medium-number-workers",
                1,
            ],
        )
    assert result.exit_code == 0
    assert "netcdf-scm: {}".format(netcdf_scm.__version__) in caplog.messages

    expected_file = join(
        OUTPUT_DIR,
        "netcdf-scm-crunched",
        "CMIP6",
        "CMIP",
        "IPSL",
        "IPSL-CM6A-LR",
        "piControl",
        "r1i1p1f1",
        "Amon",
        "tas",
        "gr",
        "v20181123",
        "netcdf-scm_tas_Amon_IPSL-CM6A-LR_piControl_r1i1p1f1_gr_284001-285912.nc",
    )

    assert isfile(expected_file)
    crunched_data = load_scmdataframe(expected_file)
    assert crunched_data.metadata["crunch_contact"] == crunch_contact
    assert crunched_data["time"].min() == dt.datetime(2840, 1, 16, 12)
    assert crunched_data["time"].max() == dt.datetime(2859, 12, 16, 12)


def test_crunching_arguments(tmpdir, caplog, test_data_marble_cmip5_dir):
    INPUT_DIR = test_data_marble_cmip5_dir
    OUTPUT_DIR = str(tmpdir)
    VAR_TO_CRUNCH = ".*fco2antt.*"
    DATA_SUB_DIR = "custom-name"
    CRUNCH_CONTACT = "test crunch contact info <email>"

    runner = CliRunner()
    with caplog.at_level("INFO"):
        result = runner.invoke(
            crunch_data,
            [
                INPUT_DIR,
                OUTPUT_DIR,
                CRUNCH_CONTACT,
                "--drs",
                "MarbleCMIP5",
                "--regexp",
                VAR_TO_CRUNCH,
                "--data-sub-dir",
                DATA_SUB_DIR,
                "-f",
                "--small-threshold",
                0,
                "--medium-threshold",
                0.5,
            ],
        )
    assert result.exit_code == 0

    assert "netcdf-scm: {}".format(netcdf_scm.__version__) in caplog.text
    assert "crunch-contact: {}".format(CRUNCH_CONTACT) in caplog.text
    assert "source: {}".format(INPUT_DIR) in caplog.text
    assert "destination: {}".format(OUTPUT_DIR) in caplog.text
    assert "drs: MarbleCMIP5" in caplog.text
    assert "regexp: {}".format(VAR_TO_CRUNCH) in caplog.text
    assert "regions: World,World|Northern Hemisphere" in caplog.text
    assert "force: True" in caplog.text
    assert "small_number_workers: 10" in caplog.text
    assert "small_threshold: 0" in caplog.text
    assert "medium_number_workers: 3" in caplog.text
    assert "medium_threshold: 0.5" in caplog.text
    assert "force_lazy_threshold: 1000" in caplog.text
    assert "cell_weights: None" in caplog.text
    assert (
        "Crunching 1 directories with greater than or equal to 0.5 million data points"
        in caplog.text
    )

    assert (
        "Making output directory: {}/custom-name".format(OUTPUT_DIR) in caplog.messages
    )

    assert "Attempting to process: ['fco2antt" in caplog.text
    assert "Attempting to process: ['tas" not in caplog.text

    assert isdir(join(OUTPUT_DIR, DATA_SUB_DIR, "cmip5"))

    out_file = join(
        OUTPUT_DIR,
        DATA_SUB_DIR,
        "cmip5",
        "1pctCO2",
        "Amon",
        "fco2antt",
        "CanESM2",
        "r1i1p1",
        "netcdf-scm_fco2antt_Amon_CanESM2_1pctCO2_r1i1p1_198001-198912.nc",
    )
    assert isfile(out_file)

    loaded = load_scmdataframe(out_file)
    assert (loaded["scenario"] == "1pctCO2").all()
    assert (loaded["climate_model"] == "CanESM2").all()
    assert (loaded["variable"] == "fco2antt").all()
    assert (
        loaded["variable_standard_name"]
        == "tendency_of_atmosphere_mass_content_of_carbon_dioxide_expressed_as_carbon_due_to_anthropogenic_emission"
    ).all()
    assert (loaded["unit"] == "kg  m^-2 s^-1").all()
    assert (loaded["member_id"] == "r1i1p1").all()
    assert (loaded["mip_era"] == "CMIP5").all()
    assert (loaded["activity_id"] == "cmip5").all()
    assert sorted(loaded["region"].unique()) == sorted(
        [
            "World",
            "World|Land",
            "World|Ocean",
            "World|Northern Hemisphere",
            "World|Northern Hemisphere|Land",
            "World|Northern Hemisphere|Ocean",
            "World|Southern Hemisphere",
            "World|Southern Hemisphere|Land",
            "World|Southern Hemisphere|Ocean",
        ]
    )
    # file is entirely zeros...
    np.testing.assert_allclose(loaded.timeseries().values, 0)

    caplog.clear()

    with caplog.at_level("INFO"):
        result_skip = runner.invoke(
            crunch_data,
            [
                INPUT_DIR,
                OUTPUT_DIR,
                "test",
                "--drs",
                "MarbleCMIP5",
                "--regexp",
                VAR_TO_CRUNCH,
                "--data-sub-dir",
                DATA_SUB_DIR,
                "--small-number-workers",
                1,
            ],
        )
    assert result_skip.exit_code == 0

    skip_str = "Skipped (already exists, not overwriting) {}".format(out_file)
    assert skip_str in caplog.text


def test_crunching_wrong_cube(tmpdir, caplog, test_data_marble_cmip5_dir):
    INPUT_DIR = test_data_marble_cmip5_dir
    OUTPUT_DIR = str(tmpdir)
    CUBE = "CMIP6Output"

    runner = CliRunner()
    with caplog.at_level("INFO"):
        result = runner.invoke(
            crunch_data, [INPUT_DIR, OUTPUT_DIR, "test", "--drs", CUBE]
        )
    assert result.exit_code  # non-zero exit code

    assert "drs: {}".format(CUBE) in caplog.text


@patch.object(
    netcdf_scm.iris_cube_wrappers._CMIPCube, "_add_time_period_from_files_in_directory"
)
def test_crunching_broken_dir(
    mock_add_time_period, tmpdir, caplog, test_data_marble_cmip5_dir
):
    mock_add_time_period.side_effect = ValueError
    INPUT_DIR = test_data_marble_cmip5_dir
    OUTPUT_DIR = str(tmpdir)
    CUBE = "CMIP6Output"

    runner = CliRunner()
    result = runner.invoke(crunch_data, [INPUT_DIR, OUTPUT_DIR, "test", "--drs", CUBE])

    assert result.exit_code  # assert failure raised
    assert "Could not calculate size of data in" in result.output, result.output


@pytest.mark.parametrize(
    "in_regions,safe,out_regions",
    [
        (["World", "World|Land"], False, ["World"]),
        (["World", "World|Ocean"], True, ["World", "World|Ocean"]),
    ],
)
def test_auto_drop_land_regions(
    in_regions, safe, out_regions, tmpdir, caplog, test_data_cmip6output_dir
):
    OUTPUT_DIR = str(tmpdir)
    crunch_contact = "join-files-test"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            crunch_data,
            [
                test_data_cmip6output_dir,
                OUTPUT_DIR,
                crunch_contact,
                "--regexp",
                ".*hfds.*",
                "--drs",
                "CMIP6Output",
                "--small-number-workers",
                1,
                "--regions",
                ",".join(in_regions),
            ],
        )
    assert result.exit_code == 0
    key_phrase = "Detected ocean data, dropping land related regions so regions to crunch are now: {}".format(
        out_regions
    )
    if not safe:
        assert key_phrase in result.stderr, result.stderr
    else:
        assert key_phrase not in result.stderr, result.stderr

    for out_file in glob(join(OUTPUT_DIR, "**", "*.nc"), recursive=True):
        res = load_scmdataframe(out_file)
        assert sorted(res["region"].unique()) == sorted(out_regions)


@pytest.mark.parametrize(
    "in_regions,safe,out_regions",
    [
        (["World", "World|Land"], True, ["World", "World|Land"]),
        (["World", "World|Ocean"], False, ["World"]),
        (
            ["World", "World|Northern Hemisphere", "World|El Nino N3.4"],
            False,
            ["World", "World|Northern Hemisphere"],
        ),
    ],
)
def test_auto_drop_ocean_regions(
    in_regions, safe, out_regions, tmpdir, caplog, test_data_cmip6output_dir
):
    OUTPUT_DIR = str(tmpdir)
    crunch_contact = "join-files-test"

    runner = CliRunner(mix_stderr=False)
    with caplog.at_level("DEBUG"):
        result = runner.invoke(
            crunch_data,
            [
                test_data_cmip6output_dir,
                OUTPUT_DIR,
                crunch_contact,
                "--regexp",
                ".*gpp.*",
                "--drs",
                "CMIP6Output",
                "--small-number-workers",
                1,
                "--regions",
                ",".join(in_regions),
            ],
        )
    assert result.exit_code == 0
    key_phrase = "Detected land data, dropping ocean related regions so regions to crunch are now: {}".format(
        out_regions
    )
    if not safe:
        assert key_phrase in result.stderr, result.stderr
    else:
        assert key_phrase not in result.stderr, result.stderr

    for out_file in glob(join(OUTPUT_DIR, "**", "*.nc"), recursive=True):
        res = load_scmdataframe(out_file)
        assert sorted(res["region"].unique()) == sorted(out_regions)


def test_crunching_cell_weighting(tmpdir, caplog, test_data_cmip6output_dir):
    crunch_contact = "join-files-test"

    runner = CliRunner(mix_stderr=False)
    run_args = [
        crunch_contact,
        "--regexp",
        ".*gpp.*",
        "--drs",
        "CMIP6Output",
        "--small-number-workers",
        1,
    ]

    out_dir_include = join(str(tmpdir), "include")
    with caplog.at_level("DEBUG"):
        result_include = runner.invoke(
            crunch_data,
            [
                test_data_cmip6output_dir,
                out_dir_include,
                "--cell-weights",
                "area-surface-fraction",
            ]
            + run_args,
        )
    assert result_include.exit_code == 0, result_include.stderr

    assert "cell_weights: area-surface-fraction" in result_include.stderr
    assert (
        "Using: <class 'netcdf_scm.weights.AreaSurfaceFractionWeightCalculator'>"
        in caplog.text
    )

    caplog.clear()
    out_dir_area_only = join(str(tmpdir), "area-only")
    with caplog.at_level("DEBUG"):
        result_area_only = runner.invoke(
            crunch_data,
            [
                test_data_cmip6output_dir,
                out_dir_area_only,
                "--cell-weights",
                "area-only",
            ]
            + run_args,
        )
    assert result_area_only.exit_code == 0, result_area_only.stderr

    assert "cell_weights: area-only" in result_area_only.stderr
    assert "Using: <class 'netcdf_scm.weights.AreaWeightCalculator'>" in caplog.text

    for out_file_include in glob(join(out_dir_include, "**", "*.nc"), recursive=True):
        res_include = load_scmdataframe(out_file_include)
        out_file_area_only = out_file_include.replace("include", "area-only")
        res_area_only = load_scmdataframe(out_file_area_only)

        for region in res_include["region"].unique():
            helper = ScmCube()
            region_key = helper._convert_region_to_area_key(region)
            res_include_effective_area = res_include.metadata[
                "{} (m**2)".format(region_key)
            ]
            res_area_only_effective_area = res_area_only.metadata[
                "{} (m**2)".format(region_key)
            ]

            assert res_area_only_effective_area >= res_include_effective_area

            ts_area_only = res_area_only.filter(region=region).timeseries()
            ts_include = res_include.filter(region=region).timeseries()
            # product of cell weighted mean times effective area should be bigger in area only
            # case
            assert (
                (
                    ts_area_only * res_area_only_effective_area
                    > ts_include * res_include_effective_area
                )
                .all()
                .all()
            )


@pytest.mark.parametrize(
    "regexp,netcdf_scm_realm,expected_include_land_surface_fraction",
    (
        (".*gpp.*", "land", True),
        (".*hfds.*", "ocean", False),
        (".*tas.*", "atmosphere", True),
    ),
)
def test_crunching_auto_assign_cell_weighting(
    test_data_cmip6output_dir,
    tmpdir,
    caplog,
    regexp,
    netcdf_scm_realm,
    expected_include_land_surface_fraction,
):
    caplog.clear()
    runner = CliRunner(mix_stderr=False)

    out_dir = str(tmpdir)
    regions_to_crunch = ["World", "World|Southern Hemisphere"]
    with caplog.at_level("DEBUG"):
        result_area_only = runner.invoke(
            crunch_data,
            [
                test_data_cmip6output_dir,
                out_dir,
                "test-auto-assign-cell-weighting",
                "--regexp",
                regexp,
                "--drs",
                "CMIP6Output",
                "--small-number-workers",
                1,
                "--regions",
                ",".join(regions_to_crunch),
            ],
        )
    assert result_area_only.exit_code == 0, result_area_only.stderr

    assert "cell_weights: None" in result_area_only.stderr
    assert "self.netcdf_scm_realm: {}".format(netcdf_scm_realm) in caplog.text
    if expected_include_land_surface_fraction:
        expected_weight_calculator = (
            "<class 'netcdf_scm.weights.AreaSurfaceFractionWeightCalculator'>"
        )
    else:
        expected_weight_calculator = "<class 'netcdf_scm.weights.AreaWeightCalculator'>"

    assert "Using: {}".format(expected_weight_calculator) in caplog.text

    for out_file in glob(join(out_dir, "**", "*.nc"), recursive=True):
        res = load_scmdataframe(out_file)
        assert res.metadata["crunch_netcdf_scm_weight_kwargs"] == json.dumps(
            {"regions": regions_to_crunch, "cell_weights": None}
        )



from abc import ABC, abstractmethod

import pytest


@pytest.mark.usefixtures(
    "picontrol_data_normalisation", "historical_data_normalisation"
)
class NormaliserIntegrationTester(ABC):
    tclass = None

    @abstractmethod
    def test_get_reference_values():
        """
        Test that reference values are retrieved as intended
        """



"""
Module for crunching raw netCDF data into netCDF-SCM netCDF files
"""
import os.path
import re

import click
import numpy as np
import tqdm

from .cli_parallel import _apply_func
from .cli_utils import (
    _find_dirs_meeting_func,
    _get_timestamp,
    _init_logging,
    _load_scm_cube,
    _make_path_if_not_exists,
    _set_crunch_contact_in_results,
)
from .definitions import NAME_COMPONENTS_SEPARATOR, OUTPUT_PREFIX
from .io import get_scmcube_helper, save_netcdf_scm_nc
from .output import OutputFileDatabase


def _crunch_data(  # pylint:disable=too-many-arguments,too-many-locals,too-many-statements
    logger,
    src,
    dst,
    crunch_contact,
    drs,
    regexp,
    regions,
    data_sub_dir,
    force,
    small_number_workers,
    small_threshold,
    medium_number_workers,
    medium_threshold,
    force_lazy_threshold,
    cell_weights,
):
    separator = NAME_COMPONENTS_SEPARATOR
    out_dir = os.path.join(dst, data_sub_dir)

    log_file = os.path.join(
        out_dir,
        "{}-crunch.log".format(_get_timestamp().replace(" ", "_").replace(":", "")),
    )
    _make_path_if_not_exists(out_dir)
    _init_logging(
        logger,
        [
            ("crunch-contact", crunch_contact),
            ("source", src),
            ("destination", out_dir),
            ("drs", drs),
            ("regexp", regexp),
            ("regions", regions),
            ("force", force),
            ("small_number_workers", small_number_workers),
            ("small_threshold", small_threshold),
            ("medium_number_workers", medium_number_workers),
            ("medium_threshold", medium_threshold),
            ("force_lazy_threshold", force_lazy_threshold),
            ("cell_weights", cell_weights),
        ],
        out_filename=log_file,
    )

    tracker = OutputFileDatabase(out_dir)
    regexp_to_match = re.compile(regexp)
    helper = get_scmcube_helper(drs)

    def keep_dir(dpath):
        if not regexp_to_match.match(dpath):
            logger.debug("Skipping (did not match regexp) %s", dpath)
            return False
        logger.info("Adding directory to queue %s", dpath)

        return True

    found_dirs, failures_dir_finding = _find_dirs_meeting_func(src, keep_dir)

    failures_calculating_data_points = False

    _get_number_data_points_kwarglist = [
        {"dpath_h": dpath, "files": files, "helper": helper}
        for dpath, files in found_dirs
    ]

    logger.info("Calculating number of points in each directory")
    dir_files_npoints = _apply_func(
        logger,
        _get_number_data_points_in_millions,
        _get_number_data_points_kwarglist,
        n_workers=small_number_workers,
        return_res=True,
    )

    dirs_to_crunch = []
    for dpath, files, no_points in tqdm.tqdm(
        dir_files_npoints, desc="Sorting directories"
    ):
        if no_points is None:
            failures_calculating_data_points = True
            try:
                helper.load_data_in_directory(dpath, process_warnings=False)
            except Exception as e:  # pylint:disable=broad-except
                logger.exception(
                    "Could not calculate size of data in %s, exception: %s", dpath, e
                )
        else:
            logger.debug("Data in %s has %s million data points", dpath, no_points)
            dirs_to_crunch.append((dpath, files, no_points))

    crunch_kwargs = {
        "logger": logger,
        "drs": drs,
        "separator": separator,
        "output_prefix": OUTPUT_PREFIX,
        "out_dir": out_dir,
        "regions": regions,
        "force": force,
        "existing_files": tracker._data,  # pylint:disable=protected-access
        "crunch_contact": crunch_contact,
        "force_lazy_threshold": force_lazy_threshold,
        "cell_weights": cell_weights,
    }

    def process_results(res):
        if res is None:
            return  # skipped crunching
        scm_timeseries_cubes, out_filepath, info = res
        logger.info("Registering %s", out_filepath)
        tracker.register(out_filepath, info)
        logger.info("Writing file to %s", out_filepath)
        save_netcdf_scm_nc(scm_timeseries_cubes, out_filepath)

    def crunch_from_list(crunch_list, n_workers=1):
        return _apply_func(
            logger,
            _crunch_files,
            crunch_list,
            common_kwarglist=crunch_kwargs,
            postprocess_func=process_results,
            n_workers=n_workers,
            style="processes",
        )

    failures_small = False
    dirs_to_crunch_small = [
        {"fnames": f, "dpath": d} for d, f, n in dirs_to_crunch if n < small_threshold
    ]
    logger.info(
        "Crunching %s directories with less than %s million data points",
        len(dirs_to_crunch_small),
        small_threshold,
    )
    if dirs_to_crunch_small:
        failures_small = crunch_from_list(
            dirs_to_crunch_small, n_workers=small_number_workers
        )

    failures_medium = False
    dirs_to_crunch_medium = [
        {"fnames": f, "dpath": d}
        for d, f, n in dirs_to_crunch
        if small_threshold <= n < medium_threshold
    ]
    logger.info(
        "Crunching %s directories with greater than or equal to %s and less than %s million data points",
        len(dirs_to_crunch_medium),
        small_threshold,
        medium_threshold,
    )
    if dirs_to_crunch_medium:
        failures_medium = crunch_from_list(
            dirs_to_crunch_medium, n_workers=medium_number_workers
        )

    failures_large = False
    dirs_to_crunch_large = [
        {"fnames": f, "dpath": d} for d, f, n in dirs_to_crunch if n > medium_threshold
    ]
    logger.info(
        "Crunching %s directories with greater than or equal to %s million data points",
        len(dirs_to_crunch_large),
        medium_threshold,
    )
    if dirs_to_crunch_large:
        failures_large = crunch_from_list(dirs_to_crunch_large, n_workers=1)

    if (
        failures_calculating_data_points
        or failures_dir_finding
        or failures_small
        or failures_medium
        or failures_large
    ):
        raise click.ClickException(
            "Some files failed to process. See {} for more details".format(log_file)
        )


def _get_number_data_points_in_millions(dpath_h, files, helper):
    try:
        helper.load_data_in_directory(dpath_h, process_warnings=False)
    except Exception:  # pylint:disable=broad-except
        return dpath_h, files, None

    data_points = np.prod(helper.cube.shape) / 10 ** 6

    return dpath_h, files, data_points


def _crunch_files(  # pylint:disable=too-many-arguments,too-many-locals
    fnames,
    dpath,
    logger=None,
    drs=None,
    separator=None,
    output_prefix=None,
    out_dir=None,
    regions=None,
    force=None,
    existing_files=None,
    crunch_contact=None,
    force_lazy_threshold=None,
    cell_weights=None,
):
    logger.info("Attempting to process: %s", fnames)
    scmcube = _load_scm_cube(drs, dpath, fnames)

    out_filename = separator.join([output_prefix, scmcube.get_data_filename()])

    outfile_dir = scmcube.get_data_directory().replace(scmcube.root_dir, out_dir)
    out_filepath = os.path.join(outfile_dir, out_filename)

    _make_path_if_not_exists(outfile_dir)

    if not force and out_filepath in existing_files:
        logger.info("Skipped (already exists, not overwriting) %s", out_filepath)
        return None

    regions = regions.split(",")
    if scmcube.netcdf_scm_realm == "ocean":
        ocean_regions = [r for r in regions if "Land" not in r]
        if set(regions) - set(ocean_regions):
            regions = ocean_regions
            logger.warning(
                "Detected ocean data, dropping land related regions so regions "
                "to crunch are now: %s",
                regions,
            )

    elif scmcube.netcdf_scm_realm == "land":
        land_regions = [
            r for r in regions if not any([ss in r for ss in ("Ocean", "El Nino")])
        ]
        if set(regions) - set(land_regions):
            regions = land_regions
            logger.warning(
                "Detected land data, dropping ocean related regions so regions "
                "to crunch are now: %s",
                regions,
            )

    ndata_points = np.prod(scmcube.cube.shape) / 10 ** 6
    lazy = ndata_points > force_lazy_threshold
    if lazy:
        logger.info(
            "Data in %s has %s million data points which is above "
            "force-lazy-threshold of %s million data points hence processing lazily",
            dpath,
            ndata_points,
            force_lazy_threshold,
        )

    results = scmcube.get_scm_timeseries_cubes(
        regions=regions, lazy=lazy, cell_weights=cell_weights,
    )
    results = _set_crunch_contact_in_results(results, crunch_contact)

    return results, out_filepath, scmcube.info


def _translate_cli_weighting_args(land_surface_fraction_weights, netcdf_scm_realm):
    if land_surface_fraction_weights == "include":
        return True

    if land_surface_fraction_weights == "area-only":
        return False

    if land_surface_fraction_weights != "guess":  # pragma: no cover # emergency valve
        raise NotImplementedError(
            "Unrecognised land_surface_fraction_weights: {}".format(
                land_surface_fraction_weights
            )
        )

    if netcdf_scm_realm == "ocean":
        return False

    # anything else e.g. land or atmosphere
    return True
